from abc import abstractproperty
from collections.abc import Iterable


from typing import List, TYPE_CHECKING
from typing import Iterable as IterableType


from pymodaq_utils.config import (BaseConfig, recursive_iterable_flattening, ConfigError,
                                  get_set_config_dir)

if TYPE_CHECKING:
    from pyqtgraph.parametertree import Parameter


def get_set_roi_path():
    """ creates and return the config folder path for managers files
    """
    return get_set_config_dir('roi_configs')


class ConfigSaverLoader:
    """ Allows to set Parameters values from previously saved one in a configuration file

    This plays the role of a cache for these Parameters

    Parameters
    ----------
    base_param: Parameter
        The parent Parameter whose children should be cached in the config file
    config: BaseConfig
        The Config object that will cache the Parameter values
    base_path: Iterable[str]
        an iterable of string defining a "category"
    """

    def __init__(self, base_param: 'Parameter', config: BaseConfig,
                 base_path: IterableType[str] = None):
        self.config = config
        if base_path is None:
            base_path = []
        self._base_path: List[str] = list(recursive_iterable_flattening(base_path))
        self._base_param = base_param

    @property
    def base_path(self):
        """ Get/Set the iterable of string defining a particular configuration to be loaded/saved"""
        return self._base_path

    @base_path.setter
    def base_path(self, path: IterableType[str]):
        self._base_path = list(recursive_iterable_flattening(path))

    @property
    def base_param(self):
        """ Get/Set the parent Parameter whose children should be saved in the config file"""
        return self._base_param

    @base_param.setter
    def base_param(self, param: 'Parameter'):
        self._base_param = param

    def load_config(self, param: 'Parameter' = None):
        from pymodaq_gui.parameter import utils as putils

        if param is None:
            param = self.base_param
        base_path = self.base_path[:]
        for child in putils.iter_children_params(param, []):
            if len(child.children()) == 0:  # means it is not a group parameter

                path = base_path + putils.get_param_path(child)[1:]

                try:
                    child.setValue(self.config(
                        *path))  # first try to load the config including the actuators name
                except ConfigError as e:
                    pass
            else:
                self.load_config(child)

    def save_config(self):
        from pymodaq_gui.parameter import utils as putils

        for param in putils.iter_children_params(self.base_param, []):
            path_param = self.base_path[:]
            path_param.extend(putils.get_param_path(param)[1:])
            try:
                self.config[tuple(path_param)] = param.value()
            except Exception as e:
                pass
        self.config.save()




from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import QObject, Signal, QLocale
import sys


class TreeLayout(QObject):
    """
    qtpy class object based on QtreeWidget
    The function populate_tree has to be used in order to populate the tree with structure as nested lists of dicts

    """
    status_sig = Signal(str)
    item_clicked_sig = Signal(object)
    item_double_clicked_sig = Signal(object)
    
    def __init__(self, parent=None, col_counts=1, labels=None):
        
        super().__init__()

        if parent is None:
            parent = QtWidgets.QWidget()
        self.parent = parent

        self.setupUi()

        self.tree.setColumnCount(col_counts)
        if labels is not None:
            self.tree.setHeaderLabels(labels)

        self.open_tree_pb.clicked.connect(self.expand_all)
        self.close_tree_pb.clicked.connect(self.collapse_all)
        self.open_tree_selected_pb.clicked.connect(self.open_tree_selection)
        
        self.tree.itemClicked.connect(self.item_clicked_sig.emit)
        self.tree.itemDoubleClicked.connect(self.item_double_clicked_sig.emit)

    def _current_text(self, col_index: int = 2):
        return self.tree.currentItem().text(col_index)

    def current_node_path(self):
        return self._current_text(2)

    def expand_all(self):
        self.tree.expandAll()

    def collapse_all(self):
        self.tree.collapseAll()

    @property
    def treewidget(self):
        return self.tree

    def setupUi(self):
        vlayout = QtWidgets.QVBoxLayout()
        hlayout = QtWidgets.QHBoxLayout()

        self.tree = CustomTree()
        vlayout.addWidget(self.tree)

        iconopen = QtGui.QIcon()
        iconopen.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/tree.png"), QtGui.QIcon.Normal,
                           QtGui.QIcon.Off)
        self.open_tree_pb = QtWidgets.QPushButton('Open Tree')
        self.open_tree_pb.setIcon(iconopen)

        iconopensel = QtGui.QIcon()
        iconopensel.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/tree.png"), QtGui.QIcon.Normal,
                              QtGui.QIcon.Off)
        self.open_tree_selected_pb = QtWidgets.QPushButton('Open Selected')
        self.open_tree_selected_pb.setIcon(iconopensel)

        iconclose = QtGui.QIcon()
        iconclose.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/CollapseAll.png"), QtGui.QIcon.Normal,
                            QtGui.QIcon.Off)
        self.close_tree_pb = QtWidgets.QPushButton('Close Tree')
        self.close_tree_pb.setIcon(iconclose)

        hlayout.addWidget(self.open_tree_pb)
        hlayout.addWidget(self.open_tree_selected_pb)
        hlayout.addWidget(self.close_tree_pb)

        vlayout.addLayout(hlayout)

        self.parent.setLayout(vlayout)

    def open_tree_selection(self):
        self.tree_open_children(self.tree.selectedIndexes()[0])

    def tree_open_children(self, item_index):
        try:
            if not (item_index.isValid()):
                return

            self.tree.expand(item_index)
        except Exception as e:
            self.status_sig.emit(str(e))

    def tree_open_parents(self, item_index):
        try:
            if not (item_index.isValid()):
                return
            flag = True
            empty = QtCore.QModelIndex()
            parent = item_index
            while flag:
                parent = parent.parent()
                if parent != empty:
                    self.tree.expand(parent)
                else:
                    flag = False
                    break
        except Exception as e:
            self.status_sig.emit(str(e))

    def populate_tree(self, data_dict):
        try:
            parents = []
            for data in data_dict:
                str_list = [str(data['name'])]
                if 'filename' in data.keys():
                    str_list.append(data['filename'])
                if 'info' in data.keys():
                    str_list.append(data['info'])
                parent = QtWidgets.QTreeWidgetItem(str_list)
                Items = self.populate_sub_tree(data['contents'])
                parent.addChildren(Items)
                parents.append(parent)

            self.tree.addTopLevelItems(parents)
        except Exception as e:
            self.status_sig.emit(str(e))

    def populate_sub_tree(self, datas):
        try:
            parents = []
            for data in datas:
                str_list = [str(data['name'])]
                if 'filename' in data.keys():
                    str_list.append(data['filename'])
                if 'info' in data.keys():
                    str_list.append(data['info'])
                parent = QtWidgets.QTreeWidgetItem(str_list)
                if 'contents' in data.keys():
                    if type(data['contents']) == list:
                        Items = self.populate_sub_tree(data['contents'])
                        parent.addChildren(Items)
                parents.append(parent)
            return parents
        except Exception as e:
            self.status_sig.emit(str(e))
            return parents


class CustomTree(QtWidgets.QTreeWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)


if __name__ == '__main__':


    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    prog = TreeLayout(Form, col_counts=2, labels=["Material", "File"])

    # example of actions to add to the tree widget in order to show a context menu
    detector_action = QtWidgets.QAction("Grab from camera", None)
    prog.tree.addAction(detector_action)
    ########################

    Form.show()
    data = [dict(name='papa', contents=[
        dict(name='fiston', contents=[dict(name='subfiston', contents='baby', filename='Cest pas sorcier')]),
        dict(name='fiston1', contents=[dict(name='subfiston', contents='baby', filename='Cest pas malin')]),
        dict(name='fiston2', contents=[dict(name='subfiston', contents='baby', filename='Cest pas normal')])]),
        dict(name='maman', contents=[dict(name='fistone', contents=[dict(name='subfistone', contents='baby')])])]
    prog.populate_tree(data)
    # filename='C:\\Data\\2019\\20190220\\Dataset_20190220_004\\Dataset_20190220_004.h5'
    # import tables
    # h5_file = tables.open_file(filename, mode = "a")
    # for node in h5_file.walk_nodes():
    #     print(node)
    # base_node = h5_file.root
    # base_tree_item, pixmap_items = h5tree_to_QTree(h5_file, base_node)
    # prog.ui.Tree.addTopLevelItem(base_tree_item)
    sys.exit(app.exec_())



from qtpy import QtWidgets, QtCore
from collections import OrderedDict
from pyqtgraph.parametertree.Parameter import ParameterItem
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem
from pyqtgraph.parametertree import Parameter
from pyqtgraph.widgets import ColorButton


class TableViewCustom(QtWidgets.QTableView):
    """
        ============== ===========================
        *Attributes**    **Type**
        *valuechanged*   instance of pyqt Signal
        *QtWidgets*      instance of QTableWidget
        ============== ===========================
    """

    valueChanged = QtCore.Signal(list)
    add_data_signal = QtCore.Signal(int)
    remove_row_signal = QtCore.Signal(int)
    load_data_signal = QtCore.Signal()
    save_data_signal = QtCore.Signal()

    def __init__(self, menu=False):
        super().__init__()
        self.setmenu(menu)

    def setmenu(self, status):
        if status:
            self.menu = QtWidgets.QMenu()
            self.menu.addAction('Add new', self.add)
            self.menu.addAction('Remove selected row', self.remove)
            self.menu.addAction('Clear all', self.clear)
            self.menu.addSeparator()
            self.menu.addAction('Load as txt', lambda: self.load_data_signal.emit())
            self.menu.addAction('Save as txt', lambda: self.save_data_signal.emit())
        else:
            self.menu = None

    def clear(self):
        self.model().clear()

    def add(self):
        self.add_data_signal.emit(self.currentIndex().row())

    def remove(self):
        self.remove_row_signal.emit(self.currentIndex().row())

    def data_has_changed(self, topleft, bottomright, roles):
        self.valueChanged.emit([topleft, bottomright, roles])

    def get_table_value(self):
        """

        """
        return self.model()

    def set_table_value(self, data_model):
        """

        """
        try:
            self.setModel(data_model)
            self.model().dataChanged.connect(self.data_has_changed)
        except Exception as e:
            pass

    def contextMenuEvent(self, event):
        if self.menu is not None:
            self.menu.exec(event.globalPos())


class TableViewParameterItem(WidgetParameterItem):
    def makeWidget(self):
        """
            Make and initialize an instance of Table_custom.

            Returns
            -------
            table : instance of Table_custom.
                The initialized table.

            See Also
            --------
            Table_custom
        """
        self.asSubItem = True
        self.hideWidget = False
        menu = False
        opts = self.param.opts
        if 'menu' in opts:
            menu = opts['menu']
        w = TableViewCustom(menu=menu)

        if 'tip' in opts:
            w.setToolTip(opts['tip'])
        w.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        #w.setMaximumHeight(200)
        # self.table.setReadOnly(self.param.opts.get('readonly', False))
        w.value = w.get_table_value
        w.setValue = w.set_table_value
        w.sigChanged = w.valueChanged
        return w

    def optsChanged(self, param, opts):
        """
            | Called when any options are changed that are not name, value, default, or limits.
            |
            | If widget is a SpinBox, pass options straight through.
            | So that only the display label is shown when visible option is toggled.

            =============== ================================== ==============================
            **Parameters**    **Type**                           **Description**
            *param*           instance of pyqtgraph parameter    the parameter to check
            *opts*            string list                        the associated options list
            =============== ================================== ==============================

            See Also
            --------
            optsChanged
        """
        # print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)

        if 'delegate' in opts:
            styledItemDelegate = QtWidgets.QStyledItemDelegate()
            styledItemDelegate.setItemEditorFactory(opts['delegate']())
            self.widget.setItemDelegate(styledItemDelegate)

        if 'menu' in opts:
            self.widget.setmenu(opts['menu'])


class TableViewParameter(Parameter):
    """
        =============== =================================
        **Attributes**    **Type**
        *itemClass*       instance of TableParameterItem
        *Parameter*       instance of pyqtgraph parameter
        =============== =================================
    """
    itemClass = TableViewParameterItem

    def setValue(self, value):
        self.opts['value'] = value
        self.sigValueChanged.emit(self, value)






from pyqtgraph.parametertree.parameterTypes.basetypes import (SimpleParameter, GroupParameter,
                                                              GroupParameterItem)
from .bool import BoolPushParameter
from .pixmap import PixmapParameter, PixmapCheckParameter
from .slide import SliderSpinBox, SliderParameter
from .led import LedPushParameter, LedParameter
from .date import DateParameter, DateTimeParameter, TimeParameter
from .list import ListParameter
from .table import TableParameter
from .tableview import TableViewParameter, TableViewCustom
from .itemselect import ItemSelectParameter
from .filedir import FileDirParameter
from .text import PlainTextPbParameter
from .numeric import NumericParameter

from pyqtgraph.parametertree.Parameter import registerParameterType, registerParameterItemType, Parameter

registerParameterType('float', NumericParameter, override=True)
registerParameterType('int',   NumericParameter, override=True)
registerParameterType('bool_push', BoolPushParameter, override=True)
registerParameterType('pixmap', PixmapParameter, override=True)
registerParameterType('pixmap_check', PixmapCheckParameter, override=True)

registerParameterType('slide', SliderParameter, override=True)

registerParameterType('led', LedParameter, override=True)
registerParameterType('led_push', LedPushParameter, override=True)
registerParameterType('date', DateParameter, override=True)
registerParameterType('date_time', DateTimeParameter, override=True)
registerParameterType('time', TimeParameter, override=True)

registerParameterType('list', ListParameter, override=True)
registerParameterType('table', TableParameter, override=True)

registerParameterType('table_view', TableViewParameter, override=True)
registerParameterType('itemselect', ItemSelectParameter, override=True)
registerParameterType('browsepath', FileDirParameter, override=True)
registerParameterType('text_pb', PlainTextPbParameter, override=True)



# -*- coding: utf-8 -*-
"""
Created the 19/10/2023

@author: Sebastien Weber
"""
import datetime

from qtpy import QtWidgets, QtCore
from qtpy.QtCore import QObject

from pymodaq_gui.parameter import ParameterTree, Parameter
from pymodaq_utils.config import Config, create_toml_from_dict


class TreeFromToml(QObject):
    """ Create a ParameterTree from a configuration file"""

    def __init__(self, config: Config = None, capitalize=True):
        super().__init__()

        if config is None:
            config = Config()
        self._config = config
        params = [{'title': 'Config path', 'name': 'config_path', 'type': 'str',
                   'value': str(self._config.config_path),
                   'readonly': True}]
        params.extend(self.dict_to_param(config.to_dict(), capitalize=capitalize))

        self.settings = Parameter.create(title='settings', name='settings', type='group',
                                         children=params)
        self.settings_tree = ParameterTree()
        self.settings_tree.setParameters(self.settings, showTop=False)

    def show_dialog(self) -> bool:

        self.dialog = QtWidgets.QDialog()
        self.dialog.setWindowTitle('Please enter new configuration values!')
        self.dialog.setLayout(QtWidgets.QVBoxLayout())
        buttonBox = QtWidgets.QDialogButtonBox(parent=self.dialog)

        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(self.dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(self.dialog.reject)

        self.dialog.layout().addWidget(self.settings_tree)
        self.dialog.layout().addWidget(buttonBox)
        self.dialog.setWindowTitle('Configuration entries')
        res = self.dialog.exec()

        if res == self.dialog.Accepted:
            with open(self._config.config_path, 'w') as f:
                config_dict = self.param_to_dict(self.settings)
                config_dict.pop('config_path')
                create_toml_from_dict(config_dict, self._config.config_path)
        return res

    @classmethod
    def param_to_dict(cls, param: Parameter) -> dict:
        config = dict()
        for child in param.children():
            if 'group' in child.opts['type']:
                config[child.name()] = cls.param_to_dict(child)
            else:
                if child.opts['type'] == 'datetime':
                    config[child.name()] = datetime.fromtimestamp(
                        child.value().toSecsSinceEpoch())  # convert QDateTime to python datetime
                elif child.opts['type'] == 'date':
                    qdt = QtCore.QDateTime()
                    qdt.setDate(child.value())
                    pdt = datetime.fromtimestamp(qdt.toSecsSinceEpoch())
                    config[child.name()] = pdt.date()
                elif child.opts['type'] == 'list':
                    config[child.name()] = child.opts['limits']
                else:
                    config[child.name()] = child.value()
        return config

    @classmethod
    def dict_to_param(cls, config: dict, capitalize=True) -> Parameter:
        params = []
        for key in config:
            if isinstance(config[key], dict):
                params.append({'title': f'{key.capitalize() if capitalize else key}:',
                               'name': key, 'type': 'group',
                               'children': cls.dict_to_param(config[key], capitalize=capitalize),
                               'expanded': 'user' in key.lower() or 'general' in key.lower()})
            else:
                param = {'title': f'{key.capitalize() if capitalize else key}:',
                         'name': key, 'value': config[key]}
                if isinstance(config[key], float):
                    param['type'] = 'float'
                elif isinstance(config[key], bool):  # placed before int because a bool is an instance of int
                    param['type'] = 'bool'
                elif isinstance(config[key], int):
                    param['type'] = 'int'
                elif isinstance(config[key], datetime.datetime):
                    param['type'] = 'datetime'
                elif isinstance(config[key], datetime.date):
                    param['type'] = 'date'
                elif isinstance(config[key], str):
                    param['type'] = 'str'
                elif isinstance(config[key], list):
                    param['type'] = 'list'
                    param['limits'] = config[key]
                    param['value'] = config[key][0]
                    param['show_pb'] = True
                params.append(param)
        return params



from qtpy import QtWidgets, QtCore
from pyqtgraph import functions as fn
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem, SimpleParameter
from pyqtgraph.parametertree import Parameter, ParameterItem
import numpy as np


class WidgetParameterItem(WidgetParameterItem):

    def updateDisplayLabel(self, value=None):
        """Update the display label to reflect the value of the parameter."""
        if value is None:
            value = self.param.value()
        self.displayLabel.setText(value.toString(self.param.opts['format']))


class DateParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        opts = self.param.opts
        w = QtWidgets.QDateEdit(QtCore.QDate(QtCore.QDate.currentDate()))
        w.setCalendarPopup(True)
        if 'format' not in opts:
            opts['format'] = 'dd/MM/yyyy'
        w.setDisplayFormat(opts['format'])

        w.sigChanged = w.dateChanged
        w.value = w.date
        w.setValue = w.setDate
        return w


class DateTimeParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        opts = self.param.opts
        w = QtWidgets.QDateTimeEdit(QtCore.QDateTime(QtCore.QDate.currentDate(),
                                                     QtCore.QTime.currentTime()))
        w.setCalendarPopup(True)
        if 'format' not in opts:
            opts['format'] = 'dd/MM/yyyy hh:mm'
        w.setDisplayFormat(opts['format'])
        w.sigChanged = w.dateTimeChanged
        w.value = w.dateTime
        w.setValue = w.setDateTime
        return w


class QTimeCustom(QtWidgets.QTimeEdit):
    def __init__(self, *args, **kwargs):
        super(QTimeCustom, self).__init__(*args, **kwargs)
        self.minutes_increment = 1
        self.timeChanged.connect(self.updateTime)

    def setTime(self, time):
        hours = time.hour()
        minutes = time.minute()

        minutes = int(np.round(minutes / self.minutes_increment) * self.minutes_increment)
        if minutes == 60:
            minutes = 0
            hours += 1

        time.setHMS(hours, minutes, 0)

        return super(QTimeCustom, self).setTime(time)

    def setMinuteIncrement(self, minutes_increment):
        self.minutes_increment = minutes_increment
        self.updateTime(self.time())

    @QtCore.Slot(QtCore.QTime)
    def updateTime(self, time):
        self.setTime(time)


class TimeParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        opts = self.param.opts
        w = QTimeCustom(QtCore.QTime(QtCore.QTime.currentTime()))
        if 'minutes_increment' in opts:
            w.setMinuteIncrement(opts['minutes_increment'])
        if 'format' not in opts:
            opts['format'] = 'hh:mm'
        w.setDisplayFormat(opts['format'])

        w.sigChanged = w.timeChanged
        w.value = w.time
        w.setValue = w.setTime
        return w

    def optsChanged(self, param, opts):
        """

        """
        super().optsChanged(param, opts)
        if 'minutes_increment' in opts:
            self.widget.setMinuteIncrement(opts['minutes_increment'])


class DateParameter(SimpleParameter):
    itemClass = DateParameterItem

    def _interpretValue(self, v):
        return QtCore.QDate(v)


class DateTimeParameter(SimpleParameter):
    itemClass = DateTimeParameterItem

    def _interpretValue(self, v):
        return QtCore.QDateTime(v)


class TimeParameter(SimpleParameter):
    itemClass = TimeParameterItem

    def _interpretValue(self, v):
        if isinstance(v, QtCore.QTime):
            return v
        elif isinstance(v, str):
            return QtCore.QTime(*eval(v.split('QTime')[1]))



import sys
import os

import numpy as np
from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtCore import Qt, QObject, Slot, Signal

from pymodaq_utils.config import get_set_local_dir, Config
from pymodaq_utils.logger import set_logger, get_module_name

from pymodaq_data.data import DataToExport, DataWithAxes
from pymodaq_data.h5modules.data_saving import DataLoader

from pymodaq_gui.utils.file_io import select_file
from pymodaq_gui.managers.parameter_manager import ParameterManager
from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.plotting.data_viewers.viewer2D_basic import Viewer2DBasic
from pymodaq_gui.plotting.items.image import UniformImageItem, SpreadImageItem
from pymodaq_gui.h5modules.browsing import browse_data
from pymodaq_gui.h5modules.saving import H5Saver
from pymodaq_gui.parameter.pymodaq_ptypes.pixmap import PixmapCheckData


local_path = get_set_local_dir()
navigator_path = os.path.join(local_path, 'navigator_temp_files')
if not os.path.isdir(navigator_path):
    os.makedirs(navigator_path)

logger = set_logger(get_module_name(__file__))

Ntick = 128
colors_red = np.array([(int(r), 0, 0) for r in np.linspace(0, 255, Ntick)])
colors_green = np.array([(0, int(g), 0) for g in np.linspace(0, 255, Ntick)])
colors_blue = np.array([(0, 0, int(b)) for b in np.linspace(0, 255, Ntick)])
config = Config()


class Navigator(ParameterManager, ActionManager, QObject):

    settings_name = 'navigator_settings'
    log_signal = Signal(str)
    sig_double_clicked = Signal(float, float)
    params = [
        {'title': 'Settings', 'name': 'settings', 'type': 'group', 'children': [
            {'title': 'Load h5:', 'name': 'Load h5', 'type': 'action'},
            {'title': 'h5 path:', 'name': 'filepath', 'type': 'str', 'value': '', 'readonly': True},
            {'title': 'Load Image:', 'name': 'Load Image', 'type': 'action'},
            {'title': 'Image path:', 'name': 'imagepath', 'type': 'str', 'value': '', 'readonly': True},
        ]},
        {'title': 'Image', 'name': 'image', 'type': 'pixmap_check'},
        {'title': 'Overlays', 'name': 'overlays', 'type': 'group', 'children': []},
    ]

    def __init__(self, parent=None, h5file_path=None):
        QObject.__init__(self)
        ParameterManager.__init__(self)
        ActionManager.__init__(self, toolbar=QtWidgets.QToolBar())
        if parent is None:
            parent = QtWidgets.QWidget()
        self.parent = parent
        self.title = 'Navigator'

        self.status_time = 1000

        self._h5saver = H5Saver()
        self.dataloader = DataLoader(self._h5saver)

        self.h5file_path = h5file_path
        self.h5saver_image = H5Saver()

        self.viewer: Viewer2DBasic = None

        self.x_range = []
        self.y_range = []

        self.next_scan_index = 0

        self.overlays = []  # %list of imageItem items displaying 2D scans info

        # self.h5module = H5BrowserUtil()

        self.setup_actions()
        self.setup_ui()
        self.connect_things()
        self.set_aspect_ratio()

        if h5file_path is not None:
            self.h5saver.open_file(h5file_path, 'a')
            self.settings.child('settings', 'filepath').setValue(h5file_path)
            self.settings.child('settings', 'Load h5').hide()
            self.list_2D_scans()
            # self.show_overlay()

    @property
    def h5saver(self):
        return self._h5saver

    @h5saver.setter
    def h5saver(self, h5saver: H5Saver):
        self._h5saver = h5saver
        self.dataloader = DataLoader(h5saver)

    def setup_actions(self):

        self.add_action('load_scan', 'Load Scan file', 'NewLayer', 'Load scan file (.h5)')
        self.add_action('load_image', 'Load Image File', 'Open_File_32', 'Load image file (.h5)')
        self.add_action('ratio', 'Viewbox Aspect Ratio', 'Zoom_1_1', 'Set viewbox aspect ratio to 1', checkable=True)
        self.add_action('move_at', 'Move at DoubleClicked', 'move_contour',
                        'When selected, double clicking on viewbox will move DAQ_Move modules',
                        checkable=True)
        self.add_action('select_all', 'Select All', 'select_all2', 'Select (show) all 2D scans on the viewer')
        self.add_action('select_none', 'Select None', 'select_none', 'Unselect (hide) all 2D scans on the viewer')
        self.add_action('histogram', 'Show Histograms', 'Histogram', 'Show (hide) histograms', checkable=True)

    def connect_things(self):
        self.connect_action('load_scan', self.load_data)
        self.connect_action('load_image', self.load_image)
        self.connect_action('ratio', self.set_aspect_ratio)
        self.connect_action('select_all', self.show_all)
        self.connect_action('select_none', self.show_none)
        self.connect_action('histogram', self.show_histo)

        self.viewer.sig_double_clicked.connect(self.move_at)
        self.settings.child('settings', 'Load h5').sigActivated.connect(self.load_data)
        self.settings.child('settings', 'Load Image').sigActivated.connect(self.load_image)

    @Slot(float, float)
    def move_at(self, posx, posy):
        if self.is_action_checked('move_at'):
            self.sig_double_clicked.emit(posx, posy)

    def show_histo(self):
        self.viewer.histo_widget.setVisible(self.is_action_checked('histogram'))

    def show_all(self):
        self.show_scans()

    def show_none(self):
        self.show_scans(False)

    def show_scans(self, show=True):
        for child in self.settings.child('overlays'):
            val = child.value()
            val.checked = show
            child.setValue(val)
            child.sigValueChanged.emit(child, val)

    def get_data_from_scan_name(self, scan_name: str) -> DataToExport:
        _data = DataToExport('All')
        self.dataloader.load_all(f'/RawData/{scan_name}', _data)
        data_2D = _data.get_data_from_Naxes(2)
        for dat in data_2D:  # convert DataWithAxes DataND to Data2D
            dat.nav_indexes = ()
        return data_2D

    def set_axes(self, dwa: DataWithAxes):
        x_axis = dwa.get_axis_from_index(0)[0]
        y_axis = dwa.get_axis_from_index(1)[0]

        self.viewer.scaled_xaxis.axis_label = x_axis.label
        self.viewer.scaled_xaxis.axis_units = x_axis.units

        self.viewer.scaled_yaxis.axis_label = y_axis.label
        self.viewer.scaled_yaxis.axis_units = y_axis.units

    def list_2D_scans(self):
        try:
            scans = self.h5saver.get_scan_groups()

            params = []
            for child in self.settings.child('overlays').children():
                self.settings.child('overlays').removeChild(child)
            ind_overlay = 0
            axes_init = False
            for scan in scans:
                data_2D = self.get_data_from_scan_name(scan.name)
                for _dat in data_2D:
                    if not axes_init:
                        self.set_axes(_dat)
                        axes_init = True
                    params.append({'name': f'Overlay{ind_overlay:03.0f}', 'type': 'pixmap_check',
                                   'value': PixmapCheckData(data=_dat[0], checked=False, path=_dat.path,
                                                            info=f'{scan.name}/{_dat.get_full_name()}')})
                    ind_overlay += 1
            self.settings.child('overlays').addChildren(params)

            for child in self.settings.child('overlays').children():
                val = child.value()
                val.checked = True
                child.setValue(val)
                child.sigValueChanged.emit(child, child.value())

        except Exception as e:
            logger.exception(str(e))

    def load_image(self):
        data, fname, node_path = browse_data(ret_all=True)

        self.h5saver_image = H5Saver()
        self.h5saver_image.open_file(fname, 'r')

        self.settings.child('settings', 'imagepath').setValue(fname)
        self.settings.child('image').setValue(PixmapCheckData(data=data[0], checked=True, path=data.path))

        self.settings.child('image').sigValueChanged.emit(
            self.settings.child('image'),
            self.settings.child('image').value())

    def load_data(self):
        self.h5file_path = str(select_file(start_path=config('data_saving', 'h5file', 'save_path'),
                                           save=False, ext='h5'))
        if self.h5file_path != '':
            self.settings.child('settings', 'filepath').setValue(self.h5file_path)
            if self.h5saver is not None:
                if self.h5saver.isopen():
                    self.h5saver.close_file()
            self.h5saver.open_file(self.h5file_path, 'a')
            self.list_2D_scans()

    def set_aspect_ratio(self):
        self.viewer.set_aspect_ratio(self.is_action_checked('ratio'))

    def add_image_data(self, dwa: DataWithAxes):
        ims = []
        histograms = self.viewer.histograms
        for ind, data in enumerate(dwa):
            if ind > 2:
                break
            if dwa.distribution.name == 'spread':
                im = SpreadImageItem()
            else:
                im = UniformImageItem()
            ims.append(im)

            im.setOpacity(1)
            self.viewer.add_image_item(im, histogram=histograms[ind])
            im.setCompositionMode(QtGui.QPainter.CompositionMode_Plus)

            if dwa.distribution.name == 'uniform':
                x_axis = dwa.get_axis_from_index(1)[0].get_data()
                y_axis = dwa.get_axis_from_index(0)[0].get_data()

                rect = QtCore.QRectF(np.min(y_axis), np.min(x_axis),
                                     (np.max(y_axis) - np.min(y_axis)),
                                     (np.max(x_axis) - np.min(x_axis)))

                im.setImage(data)
                im.setOpts(rect=rect)
            else:
                y_axis, x_axis = dwa.get_axis_from_index(0)
                im.setImage(np.vstack((x_axis, y_axis, dwa.data[0])).T)
            histograms[ind].setImageItem(im)
        return ims

    def remove_image_data(self, param):
        for overlay in self.overlays[:]:
            if param.name() in overlay['name']:
                ind = self.overlays.index(overlay)
                for image in overlay['images']:
                    self.viewer.plotitem.removeItem(image)
                self.overlays.pop(ind)

    def value_changed(self, param):

        if param.parent().name() == 'overlays':
            data: PixmapCheckData = param.value()
            if data.checked:
                dwa = self.dataloader.load_data(data.path, load_all=True)
                ims = self.add_image_data(dwa)
                self.overlays.append(dict(name=f'{param.name()}_{0}', images=ims))

            else:
                self.remove_image_data(param)
        elif param.name() == 'image':
            data: PixmapCheckData = param.value()
            if data.checked:
                dataloader = DataLoader(self.h5saver_image)
                dwa = dataloader.load_data(data.path)
                ims = self.add_image_data(dwa)
                self.overlays.append(dict(name='{:s}_{:03d}'.format(param.name(), 0), images=ims))
            else:
                self.remove_image_data(param)

    def param_deleted(self, param):
        self.remove_image_data(param)

    def setup_ui(self):
        layout = QtWidgets.QVBoxLayout()

        self.parent.setLayout(layout)
        self.parent.layout().addWidget(self.toolbar)
        sett_widget = QtWidgets.QWidget()
        self.sett_layout = QtWidgets.QHBoxLayout()
        sett_widget.setLayout(self.sett_layout)

        layout.addWidget(self.toolbar)

        splitter = QtWidgets.QSplitter(Qt.Horizontal)
        layout.addWidget(splitter)

        # set viewer area
        widg = QtWidgets.QWidget()
        self.viewer = Viewer2DBasic(widg)
        self.viewer.histogram_red.setVisible(True)
        self.viewer.histogram_green.setVisible(True)
        self.viewer.histogram_blue.setVisible(True)
        self.viewer.histogram_adaptive.setVisible(False)

        self.statusbar = QtWidgets.QStatusBar()
        self.statusbar.setMaximumHeight(25)
        layout.addWidget(self.statusbar)
        self.log_message = QtWidgets.QLabel('Initializing')
        self.statusbar.addPermanentWidget(self.log_message)

        self.sett_layout.addWidget(self.settings_tree)
        splitter.addWidget(sett_widget)
        splitter.addWidget(self.viewer.parent)

    def update_h5file(self, h5file):
        if self.h5saver is not None:
            self.h5saver.h5file = h5file
            self.dataloader = DataLoader(self.h5saver)
        self.list_2D_scans()

    def update_status(self, txt, status_time=0, log_type=None):
        """
            Show the txt message in the status bar with a delay of status_time ms.

            =============== =========== =======================
            **Parameters**    **Type**    **Description**
            *txt*             string      The message to show
            *status_time*       int         the delay of showing
            *log_type*        string      the type of the log
            =============== =========== =======================
        """
        try:
            self.ui.statusbar.showMessage(txt, status_time)
            logger.info(txt)
        except Exception as e:
            logger.exception(str(e))





from abc import ABCMeta, abstractmethod, abstractproperty
import sys
from typing import List, Tuple, Union

import numpy as np

try:
    from scipy.spatial import QhullError  # works for newer version of scipy
    from scipy.spatial import Delaunay as Triangulation
except ImportError:
    from scipy.spatial.qhull import QhullError  # works for old version of scipy
    from scipy.spatial.qhull import Delaunay as Triangulation

from qtpy import QtWidgets
from qtpy.QtCore import QObject, Slot, Signal, QRectF, QPointF

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_gui.utils.dock import DockArea, Dock
from pymodaq_gui.plotting.data_viewers.viewer1D import Viewer1D
from pymodaq_gui.plotting.utils.axes_viewer import AxesViewer
from pymodaq_gui.plotting.data_viewers.viewer2D import Viewer2D
from pymodaq_gui.plotting.data_viewers.viewer0D import Viewer0D
from pymodaq_utils import utils
from pymodaq_utils import math_utils as mutils
from pymodaq_data.data import DataRaw, Axis, DataDistribution, DataWithAxes, DataCalculated

from pymodaq_gui.plotting.data_viewers.viewer import ViewerBase
from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.managers.parameter_manager import ParameterManager
from pymodaq_data.post_treatment.process_to_scalar import DataProcessorFactory
from pymodaq_gui.managers.roi_manager import SimpleRectROI, LinearROI


logger = set_logger(get_module_name(__file__))
data_processors = DataProcessorFactory()

DEBUG_VIEWER = False


class BaseDataDisplayer(QObject):
    data_dim_signal = Signal(str)
    processor_changed = Signal(object)
    distribution: DataDistribution = abstractproperty()

    def __init__(self, viewer0D: Viewer0D, viewer1D: Viewer1D, viewer2D: Viewer2D, navigator1D: Viewer1D,
                 navigator2D: Viewer2D, axes_viewer: AxesViewer):
        super().__init__()
        self._viewer0D = viewer0D
        self._viewer1D = viewer1D
        self._viewer2D = viewer2D
        self._navigator1D = navigator1D
        self._navigator2D = navigator2D
        self._axes_viewer = axes_viewer

        self._data: DataWithAxes = None
        self._nav_limits: tuple = (0, 10, None, None)
        self._signal_at: tuple = (0, 0)

        self.triangulation = False

        self._filter_type: str = None
        self._processor = None

        self._show_nav_integration = False

    @property
    def data_shape(self):
        return self._data.shape if self._data is not None else None

    def update_filter(self, filter_type: str):
        if filter_type in self._processor.functions:
            self._filter_type = filter_type
            self.update_nav_data(*self._nav_limits)

    def update_processor(self, math_processor: DataProcessorFactory):
        self._processor = math_processor
        self.processor_changed.emit(math_processor)

    def update_data(self, data: DataRaw, force_update=False):
        if self._data is None or self._data.shape != data.shape or force_update:
            self._data = data
            self.init(data)
        else:
            self._data.data = data.data[0]

        self.data_dim_signal.emit(self._data.get_data_dimension())

        self.update_viewer_data(*self._signal_at)
        self.update_nav_data(*self._nav_limits)

    def show_nav_integration(self, show=True):
        self._show_nav_integration = show
        self.update_viewer_data(*self._signal_at)

    @abstractmethod
    def init_rois(self, data: DataRaw):
        """Init crosshairs and ROIs in viewers if needed"""
        ...

    @abstractmethod
    def init(self):
        """init viewers or postprocessing once new data are loaded"""
        ...

    @abstractmethod
    def update_viewer_data(self, **kwargs):
        """ Update the signal display depending on the position of the crosshair in the navigation panels

        """
        ...

    def updated_nav_integration(self):
        """ Means the ROI select of the 2D viewer has been moved """
        ...

    @abstractmethod
    def update_nav_data(self, x, y, width=None, height=None):
        """Display navigator data potentially postprocessed from filters in the signal viewers"""
        ...

    @abstractmethod
    def get_nav_data(self, data: DataWithAxes, x, y, width=None, height=None) -> DataWithAxes:
        """Get filtered data"""
        ...

    def update_nav_data_from_roi(self, roi: Union[SimpleRectROI, LinearROI]):
        if isinstance(roi, LinearROI):
            x, y = roi.getRegion()
            self._nav_limits = (x, y, None, None)
        elif isinstance(roi, SimpleRectROI):
            x, y = roi.pos().x(), roi.pos().y()
            width, height = roi.size().x(), roi.size().y()
            self._nav_limits = (int(x), int(y), int(width), int(height))
        self.update_nav_data(*self._nav_limits)

    @staticmethod
    def get_out_of_range_limits(x, y, width, height):
        if x < 0:
            width = width + x
            x = 0
        if y < 0:
            height = height + y
            y = 0
        return x, y, width, height

    def update_nav_indexes(self, nav_indexes: List[int]):
        self._data.nav_indexes = nav_indexes
        self.update_data(self._data, force_update=True)

    def update_nav_limits(self, x, y, width=None, height=None):
        self._nav_limits = x, y, width, height


class UniformDataDisplayer(BaseDataDisplayer):
    """Specialized object to filter and plot linearly spaced data in dedicated viewers

    Meant for any navigation axes and up to signal data dimensionality of 2 (images)
    """
    distribution = DataDistribution['uniform']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self, data: DataRaw):
        if len(data.nav_indexes) > 2 or not data.check_axes_linear():
            self._axes_viewer.set_nav_viewers(self._data.get_nav_axes_with_data())
        processor: DataProcessorFactory = data_processors
        self.update_processor(processor)

    def init_rois(self, nav_axes_limits: List[Tuple[float, float]] = None,
                  sig_axis_limits: List[Tuple[float, float]] = None):

        if len(nav_axes_limits) == 1:
            self._navigator1D.crosshair.set_crosshair_position(np.mean(nav_axes_limits[0]))

        if len(nav_axes_limits) == 2:
            self._navigator2D.crosshair.set_crosshair_position(
                *self._navigator2D.view.unscale_axis(np.mean(nav_axes_limits[1]),
                                                     np.mean(nav_axes_limits[0]))
            )
        if len(sig_axis_limits) == 1:
            self._viewer1D.roi.setPos((float(np.mean(sig_axis_limits[0]) -
                                             np.abs(np.diff(sig_axis_limits[0]))[0] / 3),
                                       float(np.mean(sig_axis_limits[0]) +
                                             np.abs(np.diff(sig_axis_limits[0]))[0] / 3))
                                      )
        if len(sig_axis_limits) == 2:
            scaled_axes = np.array(self._viewer2D.view.unscale_axis(np.array(sig_axis_limits[1]),
                                                                    np.array(sig_axis_limits[0])))

            self._viewer2D.roi.setSize(
                float(np.diff(scaled_axes[0])[0]) / 3,
                float(np.diff(scaled_axes[1])[0]) / 3)

            self._viewer2D.roi.setPos(
                float(np.mean(scaled_axes[0])) - float(np.diff(scaled_axes[0])[0]) / 6,
                float(np.mean(scaled_axes[1])) - float(np.diff(scaled_axes[1])[0]) / 6)

    def updated_nav_integration(self):
        """ Means the ROI select of the 2D viewer has been moved """
        self.update_viewer_data(*self._signal_at)

    def update_viewer_data(self, posx=0, posy=0):
        """ Update the signal display depending on the position of the crosshair in the navigation panels

        Parameters
        ----------
        posx: float
            from the 1D or 2D Navigator crosshair or from one of the navigation axis viewer (in that case
            nav_axis tells from which navigation axis the position comes from)
        posy: float
            from the 2D Navigator crosshair
        """
        self._signal_at = posx, posy
        if self._data is not None:
            try:
                if len(self._data.nav_indexes) == 0:
                    data = self._data

                elif len(self._data.nav_indexes) == 1:
                    nav_axis = self._data.axes_manager.get_nav_axes()[0]
                    if posx < nav_axis.min() or posx > nav_axis.max():
                        return
                    ind_x = nav_axis.find_index(posx)
                    logger.debug(f'Getting the data at nav index {ind_x}')
                    data: DataCalculated = self._data.inav[ind_x]
                    if self._show_nav_integration:
                        if self._navigator1D.view.is_action_checked('ROIselect'):
                            x0, x1 = self._navigator1D.view.ROIselect.getRegion()
                            ind_x0 = max(0, int(nav_axis.find_index(x0)))
                            ind_x1 = min(int(nav_axis.max()), int(nav_axis.find_index(x1)))
                            data.append(self._data.inav[ind_x0:ind_x1].mean(axis=(nav_axis.index)))
                        else:
                            data.append(self._data.mean(axis=nav_axis.index))

                elif len(self._data.nav_indexes) == 2 and self._data.check_axes_linear():
                    nav_x = self._data.axes_manager.get_nav_axes()[1]
                    nav_y = self._data.axes_manager.get_nav_axes()[0]
                    if posx < nav_x.min() or posx > nav_x.max():
                        return
                    if posy < nav_y.min() or posy > nav_y.max():
                        return
                    ind_x = nav_x.find_index(posx)
                    ind_y = nav_y.find_index(posy)
                    logger.debug(f'Getting the data at nav indexes {ind_y} and {ind_x}')
                    data = self._data.inav[ind_y, ind_x]
                    if self._show_nav_integration:
                        if self._navigator2D.view.is_action_checked('ROIselect'):
                            ind_x0 = max(0, int(self._navigator2D.view.ROIselect.x()))
                            ind_y0 = max(0, int(self._navigator2D.view.ROIselect.y()))
                            ind_x1 = min(int(nav_x.max()), ind_x0 + int(self._navigator2D.view.ROIselect.size().x()))
                            ind_y1 = min(int(nav_y.max()), ind_y0 + int(self._navigator2D.view.ROIselect.size().y()))
                            data.append(self._data.inav[ind_y0:ind_y1, ind_x0:ind_x1].mean(axis=(nav_x.index, nav_y.index)))
                        else:
                            data.append(self._data.mean(axis=(nav_x.index, nav_y.index)))
                else:
                    data = self._data.inav.__getitem__(self._axes_viewer.get_indexes())

                if len(self._data.axes_manager.sig_shape) == 0:  # means 0D data, plot on 0D viewer
                    self._viewer0D.show_data(data)

                elif len(self._data.axes_manager.sig_shape) == 1:  # means 1D data, plot on 1D viewer
                    self._viewer1D.show_data(data)

                elif len(self._data.axes_manager.sig_shape) == 2:  # means 2D data, plot on 2D viewer
                    self._viewer2D.show_data(data)
                    if DEBUG_VIEWER:
                        x, y, width, height = self.get_out_of_range_limits(*self._nav_limits)
                        _data_sig = data.isig[y: y + height, x: x + width]
                        self._debug_viewer_2D.show_data(_data_sig)

            except Exception as e:
                logger.exception(str(e))

    def update_nav_data(self, x, y, width=None, height=None):
        if self._data is not None and self._filter_type is not None and len(self._data.nav_indexes) != 0:
            nav_data = self.get_nav_data(self._data, x, y, width, height)
            if nav_data is not None:
                nav_data.nav_indexes = ()  # transform nav axes in sig axes for plotting
                if len(nav_data.shape) < 2:
                    self._navigator1D.show_data(nav_data)
                elif len(nav_data.shape) == 2 and self._data.check_axes_linear():
                    self._navigator2D.show_data(nav_data)
                else:
                    self._axes_viewer.set_nav_viewers(self._data.get_nav_axes_with_data())

    def get_nav_data(self, data: DataRaw, x, y, width=None, height=None):
        try:
            navigator_data = None
            if len(data.axes_manager.sig_shape) == 0:  # signal data is 0D
                navigator_data = data.deepcopy()

            elif len(data.axes_manager.sig_shape) == 1:  # signal data is 1D
                indx, indy = data.get_axis_from_index(data.sig_indexes[0])[0].find_indexes((x, y))
                navigator_data = self._processor.get(self._filter_type).process(data.isig[indx:indy])

            elif len(data.axes_manager.sig_shape) == 2:  # signal data is 2D
                x, y, width, height = self.get_out_of_range_limits(x, y, width, height)
                if not (width is None or height is None or width < 2 or height < 2):
                    navigator_data = self._processor.get(self._filter_type).process(data.isig[y: y + height, x: x + width])
                else:
                    navigator_data = None
            else:
                navigator_data = None

            return navigator_data

        except Exception as e:
            logger.warning('Could not compute the mathematical function')
        finally:
            return navigator_data


class SpreadDataDisplayer(BaseDataDisplayer):
    """Specialized object to filter and plot non uniformly spaced data in dedicated viewers

    Meant for any navigation axes and up to signal data dimensionality of 2 (images)
    """
    distribution = DataDistribution['spread']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.triangulation = True

    def init(self, data: DataWithAxes):
        processor = data_processors  # if len(data.axes_manager.sig_shape) > 1 else math_processors1D
        self.update_processor(processor)

    def init_rois(self, nav_axes_limits: List[Tuple[float, float]] = None,
                  sig_axis_limits: List[Tuple[float, float]] = None):
        pass

    def update_viewer_data(self, posx=0, posy=0):
        """ Update the signal display depending on the position of the crosshair in the navigation panels

        Spread data can be customly represented using:
        if signal data is 0D:
            * A viewer 1D with non-linearly spaced data points (for 1 navigation axis)
            * A viewer 2D with its SpreadImage item (for 2 navigation axis)
            * A double panel: viewer for signal data and viewer 1D for all nav axes as a function of index in the data
        otherwise:
            * A double panel: viewer for signal data and viewer 1D for all nav axes as a function of index in the data
            series

        Parameters
        ----------
        posx: float
            from the 1D or 2D Navigator crosshair or from one of the navigation axis viewer (in that case
            nav_axis tells from which navigation axis the position comes from)
        posy: float
            from the 2D Navigator crosshair
        """
        self._signal_at = posx, posy

        if self._data is not None:
            try:
                nav_axes = sorted(self._data.get_nav_axes_with_data(), key=lambda axis: axis.spread_order)

                if len(nav_axes) == 1:
                    # signal data plotted as a function of nav_axes[0] so get the index corresponding to
                    # the position posx
                    ind_nav = nav_axes[0].find_index(posx)
                    data = self._data.inav[ind_nav]

                elif len(nav_axes) == 2 and self.triangulation:
                    # signal data plotted as a function of nav_axes[0] and nav_axes[1] so get the common
                    # index corresponding to the position posx and posy
                    ind_nav, x0, y0 = mutils.find_common_index(nav_axes[0].data, nav_axes[1].data, posx, posy)
                    data = self._data.inav[ind_nav]
                else:
                    # navigation plotted as a function of index all nav_axes so get the index corresponding to
                    # the position posx
                    data = self._data.inav[int(posx)]

                if len(self._data.axes_manager.sig_shape) == 0:  # means 0D data, plot on 0D viewer
                    self._viewer0D.show_data(data)

                elif len(self._data.axes_manager.sig_shape) == 1:  # means 1D data, plot on 1D viewer
                    self._viewer1D.show_data(data)

                elif len(self._data.axes_manager.sig_shape) == 2:  # means 2D data, plot on 2D viewer
                    self._viewer2D.show_data(data)
                    if DEBUG_VIEWER:
                        x, y, width, height = self.get_out_of_range_limits(*self._nav_limits)
                        _data_sig = data.isig[y: y + height, x: x + width]
                        self._debug_viewer_2D.show_data(_data_sig)

            except Exception as e:
                logger.exception(str(e))

    def update_nav_data(self, x, y, width=None, height=None):
        if self._data is not None and self._filter_type is not None and len(self._data.nav_indexes) != 0:
            nav_data = self.get_nav_data(self._data, x, y, width, height)
            if nav_data is not None:
                nav_axes = nav_data.get_nav_axes_with_data()
                if len(nav_axes) < 2:
                    #nav_data.nav_indexes = ()
                    self._navigator1D.show_data(nav_data)
                elif len(nav_axes) == 2:
                    try:
                        Triangulation(np.array([axis.get_data() for axis in nav_data.get_nav_axes()]))
                        self._navigator2D.show_data(nav_data)
                    except QhullError as e:
                        self.triangulation = False
                        self._navigator2D.setVisible(False)
                        self._navigator1D.setVisible(True)
                        data_arrays = [axis.get_data() for axis in nav_axes]
                        labels = [axis.label for axis in nav_axes]
                        nav_data = DataCalculated('nav', data=data_arrays, labels=labels)
                        self._navigator1D.show_data(nav_data)
                else:
                    data_arrays = [axis.get_data() for axis in nav_axes]
                    labels = [axis.label for axis in nav_axes]
                    nav_data = DataCalculated('nav', data=data_arrays, labels=labels)
                    self._navigator1D.show_data(nav_data)

    def get_nav_data(self, data: DataRaw, x, y, width=None, height=None):
        try:
            navigator_data = None
            if len(data.axes_manager.sig_shape) == 0:  # signal data is 0D
                navigator_data = data

            elif len(data.axes_manager.sig_shape) == 1:  # signal data is 1D
                ind_x = data.get_axis_from_index(data.sig_indexes[0])[0].find_index(x)
                ind_y = data.get_axis_from_index(data.sig_indexes[0])[0].find_index(y)
                navigator_data = self._processor.get(self._filter_type).process(data.isig[ind_x:ind_y])

            elif len(data.axes_manager.sig_shape) == 2:  # signal data is 2D
                x, y, width, height = self.get_out_of_range_limits(x, y, width, height)
                if not (width is None or height is None or width < 2 or height < 2):
                    navigator_data = self._processor.get(self._filter_type).process(data.isig[y: y + height, x: x + width])
                else:
                    navigator_data = None
            else:
                navigator_data = None

            return navigator_data

        except Exception as e:
            logger.warning('Could not compute the mathematical function')
        finally:
            return navigator_data

    def get_nav_position(self, posx=0, posy=None):
        """
        crosshair position from the "spread" data viewer. Should return scan index where the scan was closest to posx,
        posy coordinates
        Parameters
        ----------
        posx
        posy

        See Also
        --------
        update_viewer_data
        """
        # todo adapt to new layout

        nav_axes = self.get_selected_axes()
        if len(nav_axes) != 0:
            if 'datas' in nav_axes[0]:
                datas = nav_axes[0]['datas']
                xaxis = datas[0]
                if len(datas) > 1:
                    yaxis = datas[1]
                    ind_scan = utils.find_common_index(xaxis, yaxis, posx, posy)
                else:
                    ind_scan = mutils.find_index(xaxis, posx)[0]

                self.navigator1D.ui.crosshair.set_crosshair_position(ind_scan[0])


class ViewerND(ParameterManager, ActionManager, ViewerBase):
    params = [
        {'title': 'Set data spread 1D-0D', 'name': 'set_data_spread1D0D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 1D-1D', 'name': 'set_data_spread1D1D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 1D-2D', 'name': 'set_data_spread1D2D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 2D-0D', 'name': 'set_data_spread2D0D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 2D-1D', 'name': 'set_data_spread2D1D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 2D-2D', 'name': 'set_data_spread2D2D', 'type': 'action', 'visible': False},
        {'title': 'Set data spread 3D-0D', 'name': 'set_data_spread3D0D', 'type': 'action', 'visible': False},
        {'title': 'Set data 4D', 'name': 'set_data_4D', 'type': 'action', 'visible': False},
        {'title': 'Set data 3D', 'name': 'set_data_3D', 'type': 'action', 'visible': False},
        {'title': 'Set data 2D', 'name': 'set_data_2D', 'type': 'action', 'visible': False},
        {'title': 'Set data 1D', 'name': 'set_data_1D', 'type': 'action', 'visible': False},
        {'title': 'Signal shape', 'name': 'data_shape_settings', 'type': 'group', 'children': [
            {'title': 'Initial Data shape:', 'name': 'data_shape_init', 'type': 'str', 'value': "",
             'readonly': True},
            {'title': 'Axes shape:', 'name': 'nav_axes_shapes', 'type': 'group', 'children': [],
             'readonly': True},
            {'title': 'Data shape:', 'name': 'data_shape', 'type': 'str', 'value': "", 'readonly': True},
            {'title': 'Navigator axes:', 'name': 'navigator_axes', 'type': 'itemselect'},
            {'title': 'Set Nav axes:', 'name': 'set_nav_axes', 'type': 'action', 'visible': True},
        ]},
    ]

    def __init__(self, parent: QtWidgets.QWidget = None, title=''):
        ViewerBase.__init__(self, parent, title=title)
        ActionManager.__init__(self, toolbar=QtWidgets.QToolBar())
        ParameterManager.__init__(self)

        self._area = None
        self._data = None

        self.viewer0D: Viewer0D = None
        self.viewer1D: Viewer1D = None
        self.viewer2D: Viewer2D = None
        self.navigator1D: Viewer1D = None
        self.navigator2D: Viewer2D = None
        self.axes_viewer: AxesViewer = None

        self.setup_widgets()

        self.data_displayer: BaseDataDisplayer = None

        self.setup_actions()

        self.connect_things()

        self.prepare_ui()

    def update_data_displayer(self, distribution: DataDistribution):
        if distribution.name == 'uniform':
            self.data_displayer = UniformDataDisplayer(self.viewer0D, self.viewer1D, self.viewer2D,
                                                       self.navigator1D, self.navigator2D,
                                                       self.axes_viewer)
        else:
            self.data_displayer = SpreadDataDisplayer(self.viewer0D, self.viewer1D, self.viewer2D,
                                                       self.navigator1D, self.navigator2D,
                                                       self.axes_viewer)

        self.navigator1D.crosshair.crosshair_dragged.connect(self.data_displayer.update_viewer_data)

        self.navigator1D.ROI_select_signal.connect(self.data_displayer.updated_nav_integration)

        self.navigator2D.crosshair_dragged.connect(self.data_displayer.update_viewer_data)

        self.navigator2D.ROI_select_signal.connect(self.data_displayer.updated_nav_integration)
        self.axes_viewer.navigation_changed.connect(self.data_displayer.update_viewer_data)
        self.data_displayer.data_dim_signal.connect(self.update_data_dim)

        self.viewer1D.roi.sigRegionChanged.connect(self.data_displayer.update_nav_data_from_roi)

        self.viewer2D.roi.sigRegionChanged.connect(self.data_displayer.update_nav_data_from_roi)

        self.get_action('filters').currentTextChanged.connect(self.data_displayer.update_filter)
        self.connect_action('integrate_nav', self.data_displayer.show_nav_integration)
        self.data_displayer.processor_changed.connect(self.update_filters)

    def _show_data(self, data: DataRaw, **kwargs):
        force_update = False
        self.settings.child('data_shape_settings', 'data_shape_init').setValue(str(data.shape))
        self.settings.child('data_shape_settings', 'navigator_axes').setValue(
            dict(all_items=[str(ax.index) for ax in data.axes],
                 selected=[str(ax.index) for ax in data.get_nav_axes()]))

        if (self._data is None or self._data.dim != data.dim or
                self._data.nav_indexes != data.nav_indexes):
            force_update = True
        if 'force_update' in kwargs:
            force_update = kwargs['force_update']

        if self.data_displayer is None or data.distribution != self.data_displayer.distribution:
            self.update_data_displayer(data.distribution)

        self.data_displayer.update_data(data, force_update=force_update)
        self._data = data

        if force_update:
            self.update_widget_visibility(data)
            self.data_displayer.init_rois(data.axes_limits(data.nav_indexes),
                                          data.axes_limits(data.sig_indexes))
        self.data_to_export_signal.emit(self.data_to_export)

    def set_data_test(self, data_shape='3D'):
        if 'spread' in data_shape:
            data_tri = np.load('../../../resources/triangulation_data.npy')
            axes = [Axis(data=data_tri[:, 0], index=0, label='x_axis', units='xunits', spread_order=0)]

            if 'spread2D' in data_shape or 'spread3D' in data_shape:
                axes.append(Axis(data=data_tri[:, 1], index=0, label='y_axis', units='yunits', spread_order=1))
                if data_shape == 'spread2D0D':
                    data = data_tri[:, 2]
                elif data_shape == 'spread2D1D':
                    x = np.linspace(-50, 50, 100)
                    data = np.zeros((data_tri.shape[0], len(x)))
                    for ind in range(data_tri.shape[0]):
                        data[ind, :] = data_tri[ind, 2] * mutils.gauss1D(x, 0.01*ind - 10, 20)
                    axes.append(Axis(data=x, index=1, label='sig_axis'))
                elif data_shape == 'spread2D2D':
                    x = np.linspace(-50, 50, 100)
                    y = np.linspace(-200, 200, 75)
                    data = np.zeros((data_tri.shape[0], len(y), len(x)))
                    for ind in range(data_tri.shape[0]):
                        #data[ind, :] = data_tri[ind, 2] * mutils.gauss2D(0.01*x, 0.1*ind - 20, 20, y, 0.1*ind-20, 10)
                        data[ind, :] = mutils.gauss2D(x, 10*data_tri[ind, 0], 20, y, 20*data_tri[ind, 1], 30)
                    axes.append(Axis(data=x, index=1, label='sig_axis0'))
                    axes.append(Axis(data=y, index=2, label='sig_axis1'))
                elif data_shape == 'spread3D0D':
                    if 'spread2D' in data_shape or 'spread3D' in data_shape:
                        axes.append(Axis(data=data_tri[:, 1], index=0, label='z_axis', units='zunits', spread_order=2))
                        data = data_tri[:, 2]
            
            else:
                if data_shape == 'spread1D0D':
                    data = data_tri[:, 2]
                elif data_shape == 'spread1D1D':
                    x = np.linspace(-50, 50, 100)
                    data = np.zeros((data_tri.shape[0], len(x)))
                    for ind in range(data_tri.shape[0]):
                        data[ind, :] = data_tri[ind, 2] * mutils.gauss1D(x, 0.01 * ind - 10, 20)
                    axes.append(Axis(data=x, index=1, label='sig_axis'))
                elif data_shape == 'spread1D2D':
                    x = np.linspace(-50, 50, 100)
                    y = np.linspace(-200, 200, 75)
                    data = np.zeros((data_tri.shape[0], len(y), len(x)))
                    for ind in range(data_tri.shape[0]):
                        # data[ind, :] = data_tri[ind, 2] * mutils.gauss2D(0.01*x, 0.1*ind - 20, 20, y, 0.1*ind-20, 10)
                        data[ind, :] = mutils.gauss2D(x, 10 * data_tri[ind, 0], 20, y, 20 * data_tri[ind, 1], 30)
                    axes.append(Axis(data=x, index=1, label='sig_axis0'))
                    axes.append(Axis(data=y, index=2, label='sig_axis1'))
            
            dataraw = DataRaw('NDdata', distribution='spread', dim='DataND',
                              data=[data], nav_indexes=(0, ),
                              axes=axes)
        else:
            x = mutils.linspace_step(-10, 10, 0.2)
            y = mutils.linspace_step(-30, 30, 2)
            t = mutils.linspace_step(-200, 200, 2)
            z = mutils.linspace_step(-50, 50, 0.5)
            data = np.zeros((len(y), len(x), len(t), len(z)))
            amp = mutils.gauss2D(x, 0, 5, y, 0, 4) + 0.1 * np.random.rand(len(y), len(x))
            amp = np.ones((len(y), len(x), len(t), len(z)))
            for indx in range(len(x)):
                for indy in range(len(y)):
                    data[indy, indx, :, :] = amp[indy, indx] * (
                        mutils.gauss2D(z, -50 + indx * 1, 20,
                                       t, 0 + 2 * indy, 30)
                        + np.random.rand(len(t), len(z)) / 10)

            if data_shape == '4D':
                dataraw = DataRaw('NDdata', data=data, dim='DataND', nav_indexes=[0, 1],
                                  axes=[Axis(data=y, index=0, label='y_axis', units='yunits'),
                                        Axis(data=x, index=1, label='x_axis', units='xunits'),
                                        Axis(data=t, index=2, label='t_axis', units='tunits'),
                                        Axis(data=z, index=3, label='z_axis', units='zunits')])
            elif data_shape == '3D':
                data = [np.sum(data, axis=2)]
                dataraw = DataRaw('NDdata', data=data, dim='DataND', nav_indexes=[0, 1],
                                  axes=[Axis(data=y, index=0, label='y_axis', units='yunits'),
                                        Axis(data=x, index=1, label='x_axis', units='xunits'),
                                        Axis(data=t, index=2, label='t_axis', units='tunits')])
            elif data_shape == '2D':
                data = [np.sum(data, axis=(2, 3))]
                dataraw = DataRaw('NDdata', data=data, dim='DataND', nav_indexes=[0],
                                  axes=[Axis(data=y, index=0, label='y_axis', units='yunits'),
                                        Axis(data=x, index=1, label='x_axis', units='xunits')],
                                  )
            elif data_shape == '1D':
                data = [np.sum(data, axis=(0, 1, 2))]
                dataraw = DataRaw('NDdata', data=data, dim='DataND', nav_indexes=[],
                                  axes=[Axis(data=z, index=0, label='z_axis', units='zunits')])
        self._show_data(dataraw, force_update=True)

    def update_widget_visibility(self, data: DataRaw = None,
                                 nav_indexes:Tuple[int] = None):
        if data is None:
            data = self._data
        if nav_indexes is None:
            nav_indexes = data.nav_indexes
        self.viewer0D.setVisible(len(data.shape) - len(nav_indexes) == 0)
        self.viewer1D.setVisible(len(data.shape) - len(nav_indexes) == 1)
        self.viewer2D.setVisible(len(data.shape) - len(nav_indexes) == 2)
        self.viewer1D.roi.setVisible(len(nav_indexes) != 0)
        self.viewer2D.roi.setVisible(len(nav_indexes) != 0)
        self._dock_navigation.setVisible(len(nav_indexes) != 0)
        #nav_axes = data.get_nav_axes()

        if data.distribution.name == DataDistribution.uniform:
            self.navigator1D.setVisible(len(nav_indexes) == 1)
            self.navigator2D.setVisible(len(nav_indexes) == 2 and data.check_axes_linear())
            self.axes_viewer.setVisible(len(nav_indexes) > 2 or (
                    len(nav_indexes) == 2 and not data.check_axes_linear()))
        elif data.distribution.name == DataDistribution.spread:
            self.navigator2D.setVisible(len(nav_indexes) == 2 and self.data_displayer.triangulation)
            self.navigator1D.setVisible(len(nav_indexes) == 1 or len(nav_indexes) > 2 or
                                        len(nav_indexes) == 2 and
                                        not self.data_displayer.triangulation)
        else:
            raise ValueError(f'Unknown distribution: {data.distribution.name}')

    def update_filters(self, processor: DataProcessorFactory):
        self.get_action('filters').clear()
        self.get_action('filters').addItems(processor.functions_filtered('DataND'))

    def show_settings(self, show: bool = True):
        if show:
            self.settings_tree.show()
        else:
            self.settings_tree.hide()

    def prepare_ui(self):
        self.navigator1D.setVisible(False)
        self.viewer2D.setVisible(False)
        self.navigator1D.setVisible(False)
        self.viewer2D.setVisible(False)

    def setup_actions(self):
        self.add_action('setaxes', icon_name='cartesian', checkable=True, tip='Change navigation/signal axes')
        self.add_widget('filters', QtWidgets.QComboBox, tip='Filter type to apply to signal data')
        self.add_action('integrate_nav',icon_name='integrator', checkable=True,
                        tip='Integrate the navigation data')

    def reshape_data(self):
        _nav_indexes = [int(index) for index in
                        self.settings.child('data_shape_settings', 'navigator_axes').value()['selected']]
        self.update_widget_visibility(nav_indexes=_nav_indexes)
        self.data_displayer.update_nav_indexes(_nav_indexes)

    def connect_things(self):
        self.settings.child('set_data_1D').sigActivated.connect(lambda: self.set_data_test('1D'))
        self.settings.child('set_data_2D').sigActivated.connect(lambda: self.set_data_test('2D'))
        self.settings.child('set_data_3D').sigActivated.connect(lambda: self.set_data_test('3D'))
        self.settings.child('set_data_4D').sigActivated.connect(lambda: self.set_data_test('4D'))
        self.settings.child('set_data_spread2D0D').sigActivated.connect(lambda: self.set_data_test('spread2D0D'))
        self.settings.child('set_data_spread2D1D').sigActivated.connect(lambda: self.set_data_test('spread2D1D'))
        self.settings.child('set_data_spread2D2D').sigActivated.connect(lambda: self.set_data_test('spread2D2D'))
        self.settings.child('set_data_spread1D0D').sigActivated.connect(lambda: self.set_data_test('spread1D0D'))
        self.settings.child('set_data_spread1D1D').sigActivated.connect(lambda: self.set_data_test('spread1D1D'))
        self.settings.child('set_data_spread1D2D').sigActivated.connect(lambda: self.set_data_test('spread1D2D'))
        self.settings.child('set_data_spread3D0D').sigActivated.connect(lambda: self.set_data_test('spread3D0D'))
        self.settings.child('data_shape_settings', 'set_nav_axes').sigActivated.connect(self.reshape_data)

        self.navigator1D.get_action('crosshair').trigger()
        self.connect_action('setaxes', self.show_settings)

    def setup_widgets(self):
        self.parent.setLayout(QtWidgets.QVBoxLayout())
        self.parent.layout().addWidget(self.toolbar)

        self._area = DockArea()
        self.parent.layout().addWidget(self._area)

        viewer0D_widget = QtWidgets.QWidget()
        self.viewer0D = Viewer0D(viewer0D_widget)

        viewer1D_widget = QtWidgets.QWidget()
        self.viewer1D = Viewer1D(viewer1D_widget)
        self.viewer1D.roi = LinearROI()
        self.viewer1D.view.plotitem.addItem(self.viewer1D.roi)

        viewer2D_widget = QtWidgets.QWidget()
        self.viewer2D = Viewer2D(viewer2D_widget)
        self.viewer2D.roi = SimpleRectROI(centered=True)
        self.viewer2D.view.plotitem.addItem(self.viewer2D.roi)
        
        self.viewer2D.set_action_visible('flip_ud', False)
        self.viewer2D.set_action_visible('flip_lr', False)
        self.viewer2D.set_action_visible('rotate', False)
        self.viewer2D.get_action('autolevels').trigger()

        self._dock_signal = Dock('Signal')
        self._dock_signal.addWidget(viewer0D_widget)
        self._dock_signal.addWidget(viewer1D_widget)
        self._dock_signal.addWidget(viewer2D_widget)

        navigator1D_widget = QtWidgets.QWidget()
        self.navigator1D = Viewer1D(navigator1D_widget)
        navigator2D_widget = QtWidgets.QWidget()
        self.navigator2D = Viewer2D(navigator2D_widget)
        self.navigator2D.get_action('autolevels').trigger()
        self.navigator2D.get_action('crosshair').trigger()

        nav_axes_widget = QtWidgets.QWidget()
        nav_axes_widget.setVisible(False)
        self.axes_viewer = AxesViewer(nav_axes_widget)

        self._dock_navigation = Dock('Navigation')
        self._dock_navigation.addWidget(navigator1D_widget)
        self._dock_navigation.addWidget(navigator2D_widget)
        self._dock_navigation.addWidget(nav_axes_widget)

        self._area.addDock(self._dock_navigation)
        self._area.addDock(self._dock_signal, 'right', self._dock_navigation)

    def update_data_dim(self, dim: str):
        self.settings.child('data_shape_settings', 'data_shape').setValue(dim)


def main():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = ViewerND(widget)
    for child in prog.settings.children():
        if 'set_data_' in child.name():
            child.show(True)
    prog.show_settings()

    widget.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()




from typing import Union, TYPE_CHECKING, Dict

from qtpy.QtCore import QObject, QLocale
from qtpy import QtCore, QtWidgets

from pyqtgraph.dockarea import DockArea

from pymodaq_gui.utils.dock import DockArea, Dock
from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.managers.parameter_manager import ParameterManager


class CustomApp(QObject, ActionManager, ParameterManager):
    """Base Class to ease the implementation of User Interfaces

    Inherits the MixIns ActionManager and ParameterManager classes. You have to subclass some methods and make
    concrete implementation of a given number of methods:

    * setup_actions: mandatory, see :class:`pymodaq.utils.managers.action_manager.ActionManager`
    * value_changed: non mandatory, see :class:`pymodaq.utils.managers.parameter_manager.ParameterManager`
    * child_added: non mandatory, see :class:`pymodaq.utils.managers.parameter_manager.ParameterManager`
    * param_deleted: non mandatory, see :class:`pymodaq.utils.managers.parameter_manager.ParameterManager`
    * setup_docks: mandatory
    * setup_menu: non mandatory
    * connect_things: mandatory

    Parameters
    ----------
    parent: DockArea or QtWidget

    See Also
    --------
    :class:`pymodaq.utils.managers.action_manager.ActionManager`,
    :class:`pymodaq.utils.managers.parameter_manager.ParameterManager`,
    :class:`pymodaq.utils.managers.modules_manager.ModulesManager`,
    """

    log_signal = QtCore.Signal(str)
    params = []

    def __init__(self, parent: Union[DockArea, QtWidgets.QWidget]):
        QObject.__init__(self)
        ActionManager.__init__(self)
        ParameterManager.__init__(self)

        if not isinstance(parent, DockArea):
            if not isinstance(parent, QtWidgets.QWidget):
                raise Exception('no valid parent container, expected a DockArea or a least a QWidget')

        self.parent = parent
        if isinstance(parent, DockArea):
            self.dockarea: DockArea = parent
            self.mainwindow: QtWidgets.QMainWindow = parent.parent()
        else:
            self.dockarea: DockArea = None
            self.mainwindow: QtWidgets.QMainWindow = None

        self.docks: Dict[str, Dock] = dict([])
        self.statusbar = None
        self._toolbar = QtWidgets.QToolBar()
        self._menubar: QtWidgets.QMenuBar = None

        if self.mainwindow is not None:
            self.mainwindow.addToolBar(self._toolbar)
            self._menubar = self.mainwindow.menuBar()
            self.statusbar = self.mainwindow.statusBar()

        self.set_toolbar(self._toolbar)

    def setup_ui(self):
        self.setup_docks()

        self.setup_actions()  # see ActionManager MixIn class

        self.setup_menu()

        self.connect_things()

    def setup_docks(self):
        """Mandatory method to be subclassed to setup the docks layout

        Examples
        --------
        >>>self.docks['ADock'] = gutils.Dock('ADock name')
        >>>self.dockarea.addDock(self.docks['ADock'])
        >>>self.docks['AnotherDock'] = gutils.Dock('AnotherDock name')
        >>>self.dockarea.addDock(self.docks['AnotherDock'''], 'bottom', self.docks['ADock'])

        See Also
        --------
        pyqtgraph.dockarea.Dock
        """
        raise NotImplementedError

    def setup_menu(self):
        """Non mandatory method to be subclassed in order to create a menubar

        create menu for actions contained into the self._actions, for instance:

        Examples
        --------
        >>>file_menu = self._menubar.addMenu('File')
        >>>self.affect_to('load', file_menu)
        >>>self.affect_to('save', file_menu)

        >>>file_menu.addSeparator()
        >>>self.affect_to('quit', file_menu)

        See Also
        --------
        pymodaq.utils.managers.action_manager.ActionManager
        """
        pass

    def connect_things(self):
        """Connect actions and/or other widgets signal to methods"""
        raise NotImplementedError




import os
import numpy as np
from qtpy import QtCore, QtGui
from qtpy.QtCore import Signal
import pyqtgraph as pg
import numpy as np


class Crosshair(pg.GraphicsObject):
    crosshair_dragged = Signal(float, float, name='crosshair_dragged')
    # signal used to pass crosshair position to other methods

    def __init__(self, plotitem, orientation='both', pen=None):
        super().__init__()
        
        if hasattr(plotitem, 'plotitem'):
            # in case one pass a plot or image widget
            self.plotitem = plotitem.plotitem
        else:
            self.plotitem = plotitem

        # self.crosshair_dragged[float,float].connect(self.print_pos) #exemple on how to use the crosshair_dragged signal
        if orientation == 'both':
            self.vline_visible = True
            self.hline_visible = True
        elif orientation == 'vertical':
            self.vline_visible = True
            self.hline_visible = False
        else:
            self.vline_visible = False
            self.hline_visible = True

        self.vLine = pg.InfiniteLine(angle=90, movable=True, pen=pen)
        self.hLine = pg.InfiniteLine(angle=0, movable=True, pen=pen)

        self.plotitem.addItem(self.vLine, ignoreBounds=True)
        self.plotitem.addItem(self.hLine, ignoreBounds=True)

        self.vLine.sigDragged.connect(self.update_hline)
        self.hLine.sigDragged.connect(self.update_vline)

    def show_hide_crosshair(self):
        if self.vline_visible:
            self.vLine.show()
        else:
            self.vLine.hide()
        if self.hline_visible:
            self.hLine.show()
        else:
            self.hLine.hide()

    def set_crosshair_position(self, xpos=0, ypos=0):
        self.vLine.setValue(xpos)
        self.hLine.setValue(ypos)
        self.crosshair_dragged.emit(xpos, ypos)

    def update_hline(self):
        self.hLine.sigDragged.disconnect(self.update_vline)
        p = self.vLine.getYPos() - self.vLine.cursorOffset[1]
        self.hLine.setPos(p)
        self.hLine.sigDragged.connect(self.update_vline)
        self.crosshair_dragged.emit(self.vLine.getXPos(), self.hLine.getYPos())

    def get_positions(self):
        return (self.vLine.getXPos(), self.hLine.getYPos())

    def update_vline(self):
        self.vLine.sigDragged.disconnect(self.update_hline)
        p = self.hLine.getXPos() - self.hLine.cursorOffset[0]
        self.vLine.setPos(p)
        self.vLine.sigDragged.connect(self.update_hline)
        self.crosshair_dragged.emit(self.vLine.getXPos(), self.hLine.getYPos())

    def hide(self):
        self.vLine.hide()
        self.hLine.hide()

    def show(self):
        self.show_hide_crosshair()

    def value(self):
        """
        value returns a tuple containing (x,y) positions of the crosshair
        """
        return (self.vLine.value(), self.hLine.value())

    def setVisible(self, state):
        self.hLine.setVisible(state)
        self.vLine.setVisible(state)
        if state:
            self.show_hide_crosshair()

    def isVisible(self):
        return self.hLine.isVisible()



from qtpy import QtWidgets, QtGui
from pyqtgraph.widgets.SpinBox import SpinBox


class SpinBox(SpinBox):
    """
    In case I want to add pyqtgraph spinbox functionalities
    """
    def __init__(self, *args, font_size=None, min_height=20, **kwargs):
        super().__init__(*args, **kwargs)

        if font_size is not None:
            font = QtGui.QFont()
            font.setPointSize(font_size)
            self.setFont(font)
        self.setMinimumHeight(min_height)


class QSpinBox_ro(SpinBox):
    def __init__(self, *args, readonly=True, **kwargs):
        super().__init__(*args, **kwargs)
        #self.setMaximum(100000)
        self.setReadOnly(readonly)
        self.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)



# -*- coding: utf-8 -*-
"""
Created the 19/01/2023

@author: Sebastien Weber and N Tappy
"""
# Standard imports
from collections import OrderedDict
from typing import List, Dict
from pathlib import Path
from importlib import import_module
from ..daq_utils import get_entrypoints

# 3rd party imports
import numpy as np


def register_exporter(parent_module_name: str = 'pymodaq.utils.h5modules'):
    exporters = []
    try:
        exporter_module = import_module(f'{parent_module_name}.exporters')

        exporter_path = Path(exporter_module.__path__[0])

        for file in exporter_path.iterdir():
            if file.is_file() and 'py' in file.suffix and file.stem != '__init__':
                try:
                    exporters.append(import_module(f'.{file.stem}', exporter_module.__name__))
                except ModuleNotFoundError:
                    pass
    except ModuleNotFoundError:
        pass
    finally:
        return exporters


def register_exporters() -> list:
    exporters = register_exporter('pymodaq.utils.h5modules')
    discovered_exporter_plugins = get_entrypoints('pymodaq.h5exporters')
    for entry in discovered_exporter_plugins:
        exporters.extend(register_exporter(entry.value))
    return exporters


def find_scan_node(scan_node):
    """
    utility function to find the parent node of "scan" type, meaning some of its children (DAQ_scan case)
    or co-nodes (daq_logger case) are navigation axes
    Parameters
    ----------
    scan_node: (pytables node)
        data node from where this function look for its navigation axes if any
    Returns
    -------
    node: the parent node of 'scan' type
    list: the data nodes of type 'navigation_axis' corresponding to the initial data node


    """
    try:
        while True:
            if scan_node.attrs['type'] == 'scan':
                break
            else:
                scan_node = scan_node.parent_node
        children = list(scan_node.children().values())  # for data saved using daq_scan
        children.extend([scan_node.parent_node.children()[child] for child in
                         scan_node.parent_node.children_name()])  # for data saved using the daq_logger
        nav_children = []
        for child in children:
            if 'type' in child.attrs.attrs_name:
                if child.attrs['type'] == 'navigation_axis':
                    nav_children.append(child)
        return scan_node, nav_children
    except Exception:
        return None, []


def get_h5_attributes(self, node_path):
    """
        """
    node = self.get_node(node_path)
    attrs_names = node.attrs.attrs_name
    attr_dict = OrderedDict([])
    for attr in attrs_names:
        # if attr!='settings':
        attr_dict[attr] = node.attrs[attr]

    settings = None
    scan_settings = None
    if 'settings' in attrs_names:
        if node.attrs['settings'] != '':
            settings = node.attrs['settings']

    if 'scan_settings' in attrs_names:
        if node.attrs['scan_settings'] != '':
            scan_settings = node.attrs['scan_settings']
    pixmaps = []
    for attr in attrs_names:
        if 'pixmap' in attr:
            pixmaps.append(node.attrs[attr])

    return attr_dict, settings, scan_settings, pixmaps


def get_h5_data_from_node():
    pass


def extract_axis():
    pass


def verify_axis_data_uniformity():
    pass



from typing import Union, TYPE_CHECKING, Iterable

from pymodaq_utils.enums import BaseEnum
from pyqtgraph.graphicsItems import InfiniteLine, ROI
from qtpy import QtWidgets
from qtpy.QtCore import QObject, Signal, QRectF

from pymodaq_data.data import DataToExport, DataWithAxes, DataDim, DataDistribution

from pymodaq_gui.plotting.utils.plot_utils import RoiInfo

if TYPE_CHECKING:
    from pymodaq_gui.plotting.data_viewers.viewer0D import Viewer0D
    from pymodaq_gui.plotting.data_viewers.viewer1D import Viewer1D
    from pymodaq_gui.plotting.data_viewers.viewer2D import Viewer2D
    from pymodaq_gui.plotting.data_viewers.viewerND import ViewerND


class ViewerError(Exception):
    ...


class ViewersEnum(BaseEnum):
    """enum relating a given viewer with data type"""
    Viewer0D = 'Data0D'
    Viewer1D = 'Data1D'
    Viewer2D = 'Data2D'
    ViewerND = 'DataND'
    ViewerSequential = 'DataSequential'

    def get_dim(self):
        return self.value.split('Data')[1].split('D')[0]

    def increase_dim(self, ndim: int):
        dim = self.get_dim()
        if dim != 'N':
            dim_as_int = int(dim) + ndim
            if dim_as_int > 2:
                dim = 'N'
            else:
                dim = str(dim_as_int)
        else:
            dim = 'N'
        return ViewersEnum[f'Viewer{dim}D']

    @classmethod
    def from_n_axes(cls, n_axes: int):
        if n_axes == 0:
            return ViewersEnum['Viewer0D']
        elif n_axes == 1:
            return ViewersEnum['Viewer1D']
        elif n_axes == 2:
            return ViewersEnum['Viewer2D']
        elif n_axes > 2:
            return ViewersEnum['ViewerND']

    @staticmethod
    def get_viewers_enum_from_metadata(dim: DataDim,
                                       distribution: DataDistribution,
                                       n_nav_axes: int,
                                       n_sig_indexes: int,
                                       shape_len: int,
                                       size: int) -> 'ViewersEnum':
        if dim.name == 'Data0D':
            viewer = 'Viewer0D'
        elif dim.name == 'Data1D':
            viewer = 'Viewer1D'
        elif dim.name == 'Data2D':
            viewer = 'Viewer2D'
        else:
            if distribution.name == 'uniform':
                if shape_len < 3:
                    if shape_len == 1 and size == 1:
                        viewer = 'Viewer0D'
                    elif shape_len == 1 and size > 1:
                        viewer = 'Viewer1D'
                    elif shape_len == 2:
                        viewer = 'Viewer2D'
                    else:
                        viewer = 'ViewerND'
                else:
                    viewer = 'ViewerND'
            else:
                if n_sig_indexes == 0:
                    if n_nav_axes == 1:
                        viewer = 'Viewer1D'
                    elif n_nav_axes == 2:
                        viewer = 'Viewer2D'
                    else:
                        viewer = 'ViewerND'
                else:
                    viewer = 'ViewerND'
        return ViewersEnum[viewer]

    @staticmethod
    def get_viewers_enum_from_data(dwa: DataWithAxes) -> 'ViewersEnum':
        if dwa.dim.name == 'Data0D':
            viewer = 'Viewer0D'
        elif dwa.dim.name == 'Data1D':
            viewer = 'Viewer1D'
        elif dwa.dim.name == 'Data2D':
            viewer = 'Viewer2D'
        else:
            if dwa.distribution.name == 'uniform':
                if len(dwa.shape) < 3 and dwa.check_axes_linear():
                    dwa.nav_indexes = ()
                    if len(dwa.shape) == 1 and dwa.size == 1:
                        viewer = 'Viewer0D'
                    elif len(dwa.shape) == 1 and dwa.size > 1:
                        viewer = 'Viewer1D'
                    elif len(dwa.shape) == 2:
                        viewer = 'Viewer2D'
                    else:
                        viewer = 'ViewerND'
                elif len(dwa.shape) == 1 and not dwa.check_axes_linear():
                    viewer = 'Viewer1D'
                    dwa.nav_indexes = ()
                else:
                    viewer = 'ViewerND'
            else:
                if len(dwa.sig_indexes) == 0:
                    if len(dwa.get_nav_axes()) == 1:
                        viewer = 'Viewer1D'
                    elif len(dwa.get_nav_axes()) == 2:
                        viewer = 'Viewer2D'
                    else:
                        viewer = 'ViewerND'
                else:
                    viewer = 'ViewerND'
        return ViewersEnum[viewer]


class ViewerBase(QObject):
    """Base Class for data viewers implementing all common functionalities

    Parameters
    ----------
    parent: QtWidgets.QWidget
    title: str

    Attributes
    ----------
    view: QObject
        Ui interface of the viewer

    data_to_export_signal: Signal[DataToExport]
    ROI_changed: Signal
    crosshair_dragged: Signal[float, float]
    crosshair_clicked: Signal[bool]
    sig_double_clicked: Signal[float, float]
    status_signal: Signal[str]
    """
    data_to_export_signal = Signal(DataToExport)
    _data_to_show_signal = Signal(DataWithAxes)

    ROI_changed = Signal()
    crosshair_dragged = Signal(float, float)  # Crosshair position in units of scaled top/right axes
    status_signal = Signal(str)
    crosshair_clicked = Signal(bool)
    sig_double_clicked = Signal(float, float)
    ROI_select_signal = Signal(QRectF)  # deprecated: use roi_select_signal
    roi_select_signal = Signal(RoiInfo)

    def __init__(self, parent: QtWidgets.QWidget = None, title=''):
        super().__init__()
        self.title = title if title != '' else self.__class__.__name__

        self._raw_data = None
        self.data_to_export: DataToExport = DataToExport(name=self.title)
        self.view: Union[Viewer0D, Viewer1D, Viewer2D, ViewerND] = None

        if parent is None:
            parent = QtWidgets.QWidget()
            parent.show()
        self.parent = parent

        self._display_temporary = False

    @property
    def has_action(self):
        """Convenience method"""
        if hasattr(self.view, 'has_action'):
            return self.view.has_action

    @property
    def is_action_checked(self):
        """Convenience method"""
        if hasattr(self.view, 'is_action_checked'):
            return self.view.is_action_checked

    @property
    def is_action_visible(self):
        """Convenience method"""
        if hasattr(self.view, 'is_action_visible'):
            return self.view.is_action_visible

    @property
    def set_action_checked(self):
        """Convenience method"""
        if hasattr(self.view, 'set_action_checked'):
            return self.view.set_action_checked

    @property
    def set_action_visible(self):
        """Convenience method"""
        if hasattr(self.view, 'set_action_visible'):
            return self.view.set_action_visible

    @property
    def get_action(self):
        """Convenience method"""
        if hasattr(self.view, 'get_action'):
            return self.view.get_action

    @property
    def toolbar(self):
        """Convenience property"""
        if hasattr(self.view, 'toolbar'):
            return self.view.toolbar

    @property
    def viewer_type(self):
        """str: the viewer data type see DATA_TYPES"""
        return ViewersEnum[self.__class__.__name__].value

    def show_data(self, data: DataWithAxes, **kwargs):
        """Entrypoint to display data into the viewer

        Parameters
        ----------
        data: data_mod.DataFromPlugins
        """
        if len(data.shape) > 4:
            raise ViewerError(f'Ndarray of dim: {len(data.shape)} cannot be plotted using a {self.viewer_type}')

        self.data_to_export = DataToExport(name=self.title)
        self._raw_data = data

        self._display_temporary = False

        self._show_data(data, **kwargs)

    def show_data_temp(self, data: DataWithAxes, **kwargs):
        """Entrypoint to display temporary data into the viewer

        No processed data signal is emitted from the viewer

        Parameters
        ----------
        data: data_mod.DataFromPlugins
        """
        self._display_temporary = True
        self.show_data(data, **kwargs)

    def _show_data(self, data: DataWithAxes, *args, **kwargs):
        """Specific viewers should implement it"""
        raise NotImplementedError

    def add_attributes_from_view(self):
        """Convenience function to add attributes from the view to self"""
        for attribute in self.convenience_attributes:
            if hasattr(self.view, attribute):
                setattr(self, attribute, getattr(self.view, attribute))

    def trigger_action(self, action_name: str):
        """Convenience function to trigger programmatically one of the action of the related view"""
        if self.has_action(action_name):
            self.get_action(action_name).trigger()

    def activate_roi(self, activate=True):
        """Activate the Roi manager using the corresponding action"""
        raise NotImplementedError

    def setVisible(self, show=True):
        """convenience method to show or hide the paretn widget"""
        self.parent.setVisible(show)

    @property
    def roi_target(self) -> Union[InfiniteLine.InfiniteLine, ROI.ROI]:
        """To be implemented if necessary (Viewer1D and above)"""
        return None

    def move_roi_target(self, pos: Iterable[float] = None, **kwargs):
        """move a specific read only ROI at the given position on the viewer"""
        ...

    def show_roi_target(self, show=True):
        """Show/Hide a specific read only ROI"""
        if self.roi_target is not None:
            self.roi_target.setVisible(show)





from qtpy import QtWidgets, QtCore
from pymodaq_gui.parameter.pymodaq_ptypes.itemselect import ItemSelect


class MyListWidget(QtWidgets.QListWidget):

    def sizeHint(self):
        return QtCore.QSize(super().sizeHint().width(), 25 * self.count())


def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)

    lw = MyListWidget()
    lw.setResizeMode(QtWidgets.QListView.Adjust)
    for ind in range(10):
        item = QtWidgets.QListWidgetItem(f'item{ind}')
        lw.addItem(item)
        #lw.updateGeometry()

    lw.show()

    itemsel = ItemSelect()
    itemsel.set_value(dict(all_items=[f'item{ind}' for ind in range(23)],
                           selected=[f'item{ind}' for ind in range(23)]))
    itemsel.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import os
from pathlib import Path
from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph.parametertree.Parameter import ParameterItem
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem
from pyqtgraph.parametertree import Parameter


class PlainTextWidget(QtWidgets.QWidget):
    """
        ================ ========================
        **Attributes**    **Type**
        *value_changed*   instance of pyqt Signal
        ================ ========================

        See Also
        --------
        initUI, emitsignal
    """
    value_changed = QtCore.Signal(str)

    def __init__(self):

        super().__init__()

        self.initUI()
        self.text_edit.textChanged.connect(self.emitsignal)

    def emitsignal(self):
        """
            Emit the value changed signal from the text_edit attribute.
        """
        text = self.text_edit.toPlainText()
        self.value_changed.emit(text)

    def set_value(self, txt):
        """
            Set the value of the text_edit attribute.

            =============== =========== ================================
            **Parameters**    **Type**    **Description**
            *txt*             string      the string value to be setted
            =============== =========== ================================
        """
        self.text_edit.setPlainText(txt)

    def get_value(self):
        """
            Get the value of the text_edit attribute.

            Returns
            -------
            string
                The string value of text_edit.
        """
        return self.text_edit.toPlainText()

    def initUI(self):
        """
            Init the User Interface.
        """

        self.hor_layout = QtWidgets.QHBoxLayout()
        self.text_edit = QtWidgets.QPlainTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setMaximumHeight(50)

        self.add_pb = QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.text_edit)

        verlayout = QtWidgets.QVBoxLayout()
        verlayout.addWidget(self.add_pb)
        verlayout.addStretch()
        self.hor_layout.addLayout(verlayout)
        self.hor_layout.setSpacing(0)
        self.setLayout(self.hor_layout)


class PlainTextParameterItem(WidgetParameterItem):

    def __init__(self, param, depth):
        super().__init__(param, depth)
        self.hideWidget = False
        self.subItem = QtWidgets.QTreeWidgetItem()
        self.addChild(self.subItem)

    def treeWidgetChanged(self):
        # # TODO: fix so that superclass method can be called
        # # (WidgetParameter should just natively support this style)
        # WidgetParameterItem.treeWidgetChanged(self)
        if self.treeWidget() is not None:
            self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
            self.treeWidget().setItemWidget(self.subItem, 0, self.w)

            # for now, these are copied from ParameterItem.treeWidgetChanged
            self.setHidden(not self.param.opts.get('visible', True))
            self.setExpanded(self.param.opts.get('expanded', True))

    def makeWidget(self):
        """
            Make and initialize an instance of Plain_text_pb object from parameter options dictionnary (using 'readonly' key).

            Returns
            -------
            Plain_text_pb object
                The initialized object.

            See Also
            --------
            Plain_text_pb, buttonClicked
        """
        self.w = PlainTextWidget()
        self.w.text_edit.setReadOnly(self.param.opts.get('readonly', False))
        self.w.value = self.w.get_value
        self.w.setValue = self.w.set_value
        self.w.sigChanged = self.w.value_changed
        self.w.add_pb.clicked.connect(self.buttonClicked)
        return self.w

    def buttonClicked(self):
        text, ok = QtWidgets.QInputDialog.getText(None, "Enter a value to add to the parameter",
                                                  "String value:", QtWidgets.QLineEdit.Normal)
        if ok and not (text == ""):
            self.param.setValue(self.param.value() + '\n' + text)


class PlainTextPbParameter(Parameter):
    """Editable string; displayed as large text box in the tree."""
    itemClass = PlainTextParameterItem
    sigActivated = QtCore.Signal(object)

    def activate(self):
        """
            Send the Activated signal.
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)




from qtpy import API_NAME

if API_NAME.lower() == 'pyqt5':
    from qtpy.QtCore import QVariant
else:
    def QVariant(*args):
        if len(args) > 1:
            raise TypeError('Argument of QVariant is a singlet')
        elif len(args) == 1:
            return args[0]
        else:
            return None


import os
import sys

from qtpy.QtCore import QObject, Signal, QEvent, QBuffer, QIODevice, Qt
from qtpy import QtWidgets, QtCore, QtGui

from pathlib import Path
from pymodaq_utils.config import Config
from pymodaq_utils.logger import set_logger, get_module_name
from pyqtgraph import mkQApp as mkQApppg

config = Config()
logger = set_logger(get_module_name(__file__))


def clickable(widget):
    class Filter(QObject):
        clicked = Signal()

        def eventFilter(self, obj, event):
            if obj == widget:
                if event.type() == QEvent.MouseButtonRelease:
                    if obj.rect().contains(event.pos()):
                        self.clicked.emit()
                        # The developer can opt for .emit(obj) to get the object within the slot.
                        return True
            return False

    filter = Filter(widget)
    widget.installEventFilter(filter)
    return filter.clicked


def h5tree_to_QTree(base_node, base_tree_elt=None, pixmap_items=[]):
    """
        | Convert a loaded h5 file to a QTreeWidgetItem element structure containing two columns.
        | The first is the name of the h5 current node, the second is the path of the node in the h5 structure.
        |
        | Recursive function discreasing on base_node.

        ==================   ======================================== ===============================
        **Parameters**        **Type**                                 **Description**

          *h5file*            instance class File from tables module   loaded h5 file

          *base_node*         pytables h5 node                         parent node

          *base_tree_elt*     QTreeWidgetItem                          parent QTreeWidgetItem element
        ==================   ======================================== ===============================

        Returns
        -------
        QTreeWidgetItem
            h5 structure copy converted into QtreeWidgetItem structure.

        See Also
        --------
        h5tree_to_QTree

    """

    if base_tree_elt is None:
        base_tree_elt = QtWidgets.QTreeWidgetItem([base_node.name, "", base_node.path])
    for node_name, node in base_node.children().items():
        child = QtWidgets.QTreeWidgetItem([node_name, "", node.path])
        if 'pixmap' in node.attrs.attrs_name:
            pixmap_items.append(dict(node=node, item=child))
        klass = node.attrs['CLASS']
        if klass == 'GROUP':
            h5tree_to_QTree(node, child, pixmap_items)

        base_tree_elt.addChild(child)
    return base_tree_elt, pixmap_items


def set_enable_recursive(children, enable=False):
    """Apply the enable state on all children widgets, do it recursively

    Parameters
    ----------
    children: (list) elements children ofa pyqt5 element
    enable: (bool) set enabled state (True) of all children widgets
    """
    for child in children:
        if not children:
            return
        elif isinstance(child, QtWidgets.QSpinBox) or isinstance(child, QtWidgets.QComboBox) or \
                isinstance(child, QtWidgets.QPushButton) or isinstance(child, QtWidgets.QListWidget):
            child.setEnabled(enable)
        else:
            set_enable_recursive(child.children(), enable)


def widget_to_png_to_bytes(widget, keep_aspect=True, width=200, height=100):
    """
    Renders the widget content in a png format as a bytes string
    Parameters
    ----------
    widget: (QWidget) the widget to render
    keep_aspect: (bool) if True use width and the widget aspect ratio to calculate the height
                        if False use set values of width and height to produce the png
    width: (int) the rendered width of the png
    height: (int) the rendered width of the png

    Returns
    -------
    binary string

    """
    png = widget.grab().toImage()
    wwidth = widget.width()
    wheight = widget.height()
    if keep_aspect:
        height = width * wheight / wwidth

    png = png.scaled(int(width), int(height), QtCore.Qt.KeepAspectRatio)
    buffer = QtCore.QBuffer()
    buffer.open(QtCore.QIODevice.WriteOnly)
    png.save(buffer, "png")
    return buffer.data().data()


def pngbinary2Qlabel(databinary, scale_height: int = None):
    buff = QBuffer()
    buff.open(QIODevice.WriteOnly)
    buff.write(databinary)
    dat = buff.data()
    pixmap = QtGui.QPixmap()
    pixmap.loadFromData(dat, 'PNG')
    if scale_height is not None and isinstance(scale_height, int):
        pixmap = pixmap.scaledToHeight(scale_height)
    label = QtWidgets.QLabel()
    label.setPixmap(pixmap)
    return label


def start_qapplication() -> QtWidgets.QApplication:
    app = QtWidgets.QApplication(sys.argv)
    if config('style', 'darkstyle'):
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api=os.environ['QT_API']))
    return app


def mkQApp(name: str):
    app = mkQApppg(name)
    if config('style', 'darkstyle'):
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api=os.environ['QT_API']))
    return app


def exec():
    app = mkQApp('a name')
    return app.exec() if hasattr(app, 'exec') else app.exec_()



import warnings

from pymodaq_utils.warnings import deprecation_msg
from multipledispatch import dispatch
from typing import Union, Callable, List

from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtWidgets import QAction

from pymodaq_gui.QtDesigner_Ressources import QtDesigner_ressources_rc
from pathlib import Path


def create_icon(icon_name: str):
    icon = QtGui.QIcon()
    if Path(icon_name).is_file():
        icon.addPixmap(QtGui.QPixmap(icon_name), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
    else:
        icon.addPixmap(QtGui.QPixmap(f":/icons/Icon_Library/{icon_name}.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
    return icon


class QAction(QAction):
    """
    QAction subclass to mimic signals as pushbuttons. Done to be sure of backcompatibility
    when I moved from pushbuttons to QAction
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def click(self):
        deprecation_msg("click for PyMoDAQ's QAction is deprecated, use *trigger*",
                        stacklevel=3)
        self.trigger()

    @property
    def clicked(self):
        deprecation_msg("clicked for PyMoDAQ's QAction is deprecated, use *trigger*",
                        stacklevel=3)
        return self.triggered

    def connect_to(self, slot):
        self.triggered.connect(slot)

    def set_icon(self, icon_name: str):
        self.setIcon(create_icon(icon_name))


def addaction(name: str = '', icon_name: str = '', tip='', checkable=False, checked=False,
              slot: Callable = None, toolbar: QtWidgets.QToolBar = None,
              menu: QtWidgets.QMenu = None, visible=True, shortcut=None,
              enabled=True):
    """Create a new action and add it eventually to a toolbar and a menu

    Parameters
    ----------
    name: str
        Displayed name if should be displayed (for instance in menus)
    icon_name: str
        png file name to produce the icon
    tip: str
        a tooltip to be displayed when hovering above the action
    checkable: bool
        set the checkable state of the action
    checked: bool
        set the current state of the action
    slot: callable
        Method or function that will be called when the action is triggered
    toolbar: QToolBar
        a toolbar where action should be added.
    menu: QMenu
        a menu where action should be added.
    visible: bool
        display or not the action in the toolbar/menu
    shortcut: str
        a string defining a shortcut for this action
    enabled: bool
        set the enabled state
    """
    if icon_name != '':
        action = QAction(create_icon(icon_name), name, None)
    else:
        action = QAction(name)

    if slot is not None:
        action.connect_to(slot)
    action.setCheckable(checkable)
    if checkable:
        action.setChecked(checked)
    action.setToolTip(tip)
    if toolbar is not None:
        toolbar.addAction(action)
    if menu is not None:
        menu.addAction(action)
    if shortcut is not None:
        action.setShortcut(shortcut)
    action.setVisible(visible)
    action.setEnabled(enabled)
    return action


def addwidget(klass: Union[str, QtWidgets.QWidget], *args, tip='', toolbar: QtWidgets.QToolBar = None, visible=True,
              signal_str=None, slot: Callable=None, setters = {}, **kwargs):
    """Create and eventually add a widget to a toolbar

    Parameters
    ----------
    klass: str or QWidget
        should be a custom widget class or the name of a standard widget of QWidgets
    args: list
     variable arguments passed as is to the widget constructor
    tip: str
        a tooltip to be displayed when hovering above the widget
    toolbar: QToolBar
        a toolbar where the widget should be added.
    visible: bool
        display or not the action in the toolbar/menu
    signal_str: str
        an attribute of type Signal of the widget
    slot: Callable
        a callable connected to the signal
    kwargs: dict
        variable named arguments used as is in the widget constructor
    setters: dict
        method/value pair of the widget (for instance setMaximumWidth)
    Returns
    -------
    QtWidgets.QWidget
    """
    if isinstance(klass, str):
        if hasattr(QtWidgets, klass):
            widget: QtWidgets.QWidget = getattr(QtWidgets, klass)(*args)
        else:
            return None
    else:
        try:
            widget = klass(*args, **kwargs)
        except:
            return None
    widget.setVisible(visible)
    widget.setToolTip(tip)
    if toolbar is not None:
        toolbar.addWidget(widget)
    if isinstance(signal_str, str) and slot is not None:
        if hasattr(widget, signal_str):
            getattr(widget, signal_str).connect(slot)

    for setter in setters:
        if hasattr(widget, setter):
            getattr(widget, setter)(setters[setter])

    return widget


class ActionManager:
    """MixIn Class to be used by all UserInterface to manage their QActions and the action they are connected to

    Parameters
    ----------
    toolbar: QToolbar, optional
        The toolbar to use as default
    menu: QMenu, option
        The menu to use as default
    """
    def __init__(self, toolbar=None, menu=None):
        self._actions = dict([])
        self._toolbar = toolbar
        self._menu = menu

        #self.setup_actions()

    def setup_actions(self):
        """Method where to create actions to be subclassed. Mandatory

        Examples
        --------
        >>> self.add_action('Quit', 'close2', "Quit program")
        >>> self.add_action('Grab', 'camera', "Grab from camera", checkable=True)
        >>> self.add_action('Load', 'Open', "Load target file (.h5, .png, .jpg) or data from camera", checkable=False)
        >>> self.add_action('Save', 'SaveAs', "Save current data", checkable=False)

        See Also
        --------
        ActionManager.add_action
        """
        raise NotImplementedError(f'You have to define actions here in the following form:'
                                  f'{self.setup_actions.__doc__}')

    def add_action(self, short_name: str = '', name: str = '', icon_name: str = '', tip='',
                   checkable=False,
                   checked=False, toolbar=None, menu=None,
                   visible=True, shortcut=None, auto_toolbar=True, auto_menu=True,
                   enabled=True):
        """Create a new action and add it to toolbar and menu

        Parameters
        ----------
        short_name: str
            the name as referenced in the dict self.actions
        name: str
            Displayed name if should be displayed in
        icon_name: str
            png file name to produce the icon
        tip: str
            a tooltip to be displayed when hovering above the action
        checkable: bool
            set the checkable state of the action
        checked: bool
            set the current state of the action            
        toolbar: QToolBar
            a toolbar where action should be added. Actions can also be added later see *affect_to*
        menu: QMenu
            a menu where action should be added. Actions can also be added later see *affect_to*
        visible: bool
            display or not the action in the toolbar/menu
        auto_toolbar: bool
            if True add this action to the defined toolbar
        auto_menu: bool
            if True add this action to the defined menu
        enabled: bool
            set the enabled state of this action
        See Also
        --------
        affect_to, pymodaq.resources.QtDesigner_Ressources.Icon_Library,
        pymodaq.utils.managers.action_manager.add_action
        """
        if auto_toolbar:
            if toolbar is None:
                toolbar = self._toolbar
        if auto_menu:
            if menu is None:
                menu = self._menu
        self._actions[short_name] = addaction(name, icon_name, tip, checkable=checkable,
                                              checked=checked, toolbar=toolbar, menu=menu,
                                              visible=visible, shortcut=shortcut, enabled=enabled)

    def add_widget(self, short_name, klass: Union[str, QtWidgets.QWidget], *args, tip='',
                   toolbar: QtWidgets.QToolBar = None, visible=True, signal_str=None, slot: Callable=None, **kwargs):
        """Create and add a widget to a toolbar

        Parameters
        ----------
        short_name: str
            the name as referenced in the dict self.actions
        klass: str or QWidget
            should be a custom widget class or the name of a standard widget of QWidgets
        args: list
         variable arguments passed as is to the widget constructor
        tip: str
            a tooltip to be displayed when hovering above the widget
        toolbar: QToolBar
            a toolbar where the widget should be added.
        visible: bool
            display or not the action in the toolbar/menu
        signal_str: str
            an attribute of type Signal of the widget
        slot: Callable
            a callable connected to the signal
        kwargs: dict
            variable named arguments passed as is to the widget constructor
        Returns
        -------
        QtWidgets.QWidget
        """
        if toolbar is None:
            toolbar = self._toolbar
        widget = addwidget(klass, *args, tip=tip, toolbar=toolbar, visible=visible, signal_str=signal_str,
                           slot=slot, **kwargs)
        if widget is not None:
            self._actions[short_name] = widget
        else:
            warnings.warn(UserWarning(f'Impossible to add the widget {short_name} and type {klass} to the toolbar'))

    def set_toolbar(self, toolbar):
        """affect a toolbar to self

        Parameters
        ----------
        toolbar:
            QtWidgets.QToolBar
        """
        self._toolbar = toolbar

    def set_menu(self, menu):
        """affect a menu to self

        Parameters
        ----------
        menu:
            QtWidgets.QMenu
        """
        self._menu = menu

    def set_action_text(self, action_name: str, text: str):
        """Convenience method to set the displayed text on an action

        Parameters
        ----------
        action_name: str
            The action name as defined in setup_actions
        text: str
            The text to display
        """
        self.get_action(action_name).setText(text)

    @property
    def actions(self) -> List[QAction]:
        return list(self._actions.values())

    def get_action(self, name) -> QAction:
        """Getter of a given action

        Parameters
        ----------
        name: str
            The action name as defined in setup_actions

        Returns
        -------
        QAction
        """
        if self.has_action(name):
            return self._actions[name]
        else:
            raise KeyError(f'The action with name: {name} is not referenced'
                           f' in the view actions: {self._actions.keys()}')

    def has_action(self, action_name) -> bool:
        """Check if an action has been defined
        Parameters
        ----------
        action_name: str
            The action name as defined in setup_actions

        Returns
        -------
        bool: True if the action exists, False otherwise
        """
        return action_name in self._actions

    @property
    def toolbar(self):
        """Get the default toolbar"""
        return self._toolbar

    @property
    def menu(self):
        """Get the default menu"""
        return self._menu

    def affect_to(self, action_name, obj: Union[QtWidgets.QToolBar, QtWidgets.QMenu]):
        """Affect action to an object either a toolbar or a menu

        Parameters
        ----------
        action_name: str
            The action name as defined in setup_actions
        obj: QToolbar or QMenu
            The object where to add the action
        """
        if isinstance(obj, QtWidgets.QToolBar) or isinstance(obj, QtWidgets.QMenu):
            obj.addAction(self._actions[action_name])

    def connect_action(self, name, slot, connect=True, signal_name=''):
        """Connect (or disconnect) the action referenced by name to the given slot

        Parameters
        ----------
        name: str
            key of the action as referenced in the self._actions dict
        slot: method
            a method/function
        connect: bool
            if True connect the trigger signal of the action to the defined slot else disconnect it
        signal_name: str
            try to use it as a signal (for widgets added...) otherwise use the *triggered* signal
        """
        signal = 'triggered'
        if name in self._actions:
            if hasattr(self._actions[name], signal_name):
                signal = signal_name
            if connect:
                getattr(self._actions[name], signal).connect(slot)
            else:
                try:
                    getattr(self._actions[name], signal).disconnect()
                except (TypeError,) as e:
                    pass  # the action was not connected
        else:
            raise KeyError(f'The action with name: {name} is not referenced'
                           f' in the view actions: {self._actions.keys()}')

    @dispatch(str)
    def is_action_visible(self, action_name: str):
        """Check the visibility of a given action or the list of an action"""
        if action_name in self._actions:
            return self._actions[action_name].isVisible()
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list)
    def is_action_visible(self, actions_name: list):
        """Check the visibility of a given action or the list of an action"""
        isvisible = False
        for action_name in actions_name:
            isvisible = isvisible and self.is_action_visible(action_name)
        return isvisible

    @dispatch(str)
    def is_action_checked(self, action_name: str):
        """Get the CheckState of a given action or a list of actions"""
        if action_name in self._actions:
            return self._actions[action_name].isChecked()
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list)
    def is_action_checked(self, actions_name: list):
        """Get the CheckState of a given action or a list of actions"""
        ischecked = False
        for action_name in actions_name:
            ischecked = ischecked and self.is_action_checked(action_name)
        return ischecked

    @dispatch(str, bool)
    def set_action_visible(self, action_name: str, visible=True):
        """Set the visibility of a given action or a list of an action"""
        if action_name in self._actions:
            self._actions[action_name].setVisible(visible)
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list, bool)
    def set_action_visible(self, actions_name: list, visible=True):
        """Set the visibility of a given action or a list of an action"""
        for action_name in actions_name:
            self.set_action_visible(action_name, visible)

    @dispatch(str, bool)
    def set_action_checked(self, action_name: str, checked=True):
        """Set the CheckedState of a given action or a list of actions"""
        if action_name in self._actions:
            self._actions[action_name].setChecked(checked)
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list, bool)
    def set_action_checked(self, actions_name: list, checked=True):
        """Set the CheckedState of a given action or a list of actions"""
        for action_name in actions_name:
            self.set_action_checked(action_name, checked)

    @dispatch(str, bool)
    def set_action_enabled(self, action_name: str, enabled=True):
        """Set the EnabledState of a given action or a list of actions"""
        if action_name in self._actions:
            self._actions[action_name].setEnabled(enabled)
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list, bool)
    def set_action_enabled(self, actions_name: list, enabled=True):
        """Set the EnabledState of a given action or a list of actions"""
        for action_name in actions_name:
            self.set_action_enabled(action_name, enabled)

    @dispatch(str)
    def is_action_enabled(self, action_name: str):
        """Get the EnabledState of a given action or a list of actions"""
        if action_name in self._actions:
            return self._actions[action_name].isEnabled()
        else:
            raise KeyError(f'The action with name: {action_name} is not referenced'
                           f' in the actions list: {self._actions}')

    @dispatch(list)
    def is_action_checked(self, actions_name: list):
        """Get the EnabledState of a given action or a list of actions"""
        is_enabled = False
        for action_name in actions_name:
            is_enabled = is_enabled and self.is_action_enabled(action_name)
        return is_enabled




from pymodaq_gui.utils.widgets.table import SpinBoxDelegate
from  pymodaq_gui.parameter.utils import get_widget_from_tree
from pymodaq_gui.plotting.data_viewers.viewer2D import Viewer2D
from pymodaq.utils import gui_utils as gutils
from pymodaq_gui.utils.dock import Dock
from pymodaq_gui.parameter import ParameterTree, Parameter
from pymodaq_gui.parameter.pymodaq_ptypes.tableview import TableViewCustom
from pymodaq.utils.scanner.scanners.tabular import TableModelTabular
from qtpy.QtCore import QObject, Qt, Slot
from qtpy import QtWidgets


class ViewerPointList(QObject):
    def __init__(self, area):
        super().__init__()
        self.area = area
        self.viewer: Viewer2D = None

        self.set_viewer()
        self.set_point_list()
        self.viewer.sig_double_clicked.connect(self.double_click_action)

    @Slot(float, float)
    def double_click_action(self, posx, posy):
        xs, ys = self.viewer.view.unscale_axis(posx, posy)
        data_at = self.viewer.view.get_data_at('red', (xs, ys))
        if data_at is not None:
            self.table_model.add_data(self.table_view.currentIndex().row() + 1, [posx, posy, data_at])

    def show_data(self, data):
        self.viewer.show_data(data)

    def setXaxis(self, xaxis):
        self.viewer.x_axis = xaxis

    def setYaxis(self, yaxis):
        self.viewer.y_axis = yaxis

    def set_viewer(self):
        dock_viewer = Dock('Viewer2D')
        self.area.addDock(dock_viewer, 'right')
        widget = QtWidgets.QWidget()
        self.viewer = Viewer2D(widget)
        dock_viewer.addWidget(widget)

    def set_point_list(self):
        dock_list = Dock('List of points')
        self.area.addDock(dock_list, 'right')
        params = [{'title': 'Positions', 'name': 'tabular_table', 'type': 'table_view',
                   'delegate': SpinBoxDelegate, 'menu': True}, ]
        self.settings_tree = ParameterTree()
        self.settings = Parameter.create(name='settings', title='Settings', type='group', children=params)
        self.settings_tree.setParameters(self.settings, showTop=False)
        dock_list.addWidget(self.settings_tree)

        init_data = [[0., 0., 0.]]
        self.table_model = TableModelTabular(init_data, ['x', 'y', 'data'])
        self.table_view = get_widget_from_tree(self.settings_tree, TableViewCustom)[0]
        self.settings.child(('tabular_table')).setValue(self.table_model)

        self.table_view.horizontalHeader().ResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        styledItemDelegate = QtWidgets.QStyledItemDelegate()
        styledItemDelegate.setItemEditorFactory(SpinBoxDelegate())
        self.table_view.setItemDelegate(styledItemDelegate)

        self.table_view.setDragEnabled(True)
        self.table_view.setDropIndicatorShown(True)
        self.table_view.setAcceptDrops(True)
        self.table_view.viewport().setAcceptDrops(True)
        self.table_view.setDefaultDropAction(Qt.MoveAction)
        self.table_view.setDragDropMode(QtWidgets.QTableView.InternalMove)
        self.table_view.setDragDropOverwriteMode(False)

        self.table_view.add_data_signal[int].connect(self.table_model.add_data)
        self.table_view.remove_row_signal[int].connect(self.table_model.remove_data)
        self.table_view.load_data_signal.connect(self.table_model.load_txt)
        self.table_view.save_data_signal.connect(self.table_model.save_txt)


if __name__ == '__main__':
    from pymodaq.utils.gui_utils import DockArea
    from pymodaq.utils.data import Axis, DataFromPlugins
    import sys
    import numpy as np

    app = QtWidgets.QApplication(sys.argv)
    area = DockArea()
    win = QtWidgets.QMainWindow()
    win.setCentralWidget(area)
    viewer = ViewerPointList(area)

    Nx = 100
    Ny = 200
    x = (np.linspace(0, Nx - 1, Nx) + 100) / 2
    y = (np.linspace(0, Ny - 1, Ny) - 10) * 2
    from pymodaq.utils.math_utils import gauss2D

    data_red = 3 * gauss2D(x, np.mean(x), (np.max(x)-np.min(x)) / 5, y, np.mean(y), (np.max(y)-np.min(y)) / 5, 1)
    data_red += np.random.random(data_red.shape)
    data_to_plot = DataFromPlugins(name='mydata', distribution='uniform', data=[data_red],
                                   axes=[Axis('xaxis', units='xpxl', data=x, index=1),
                                         Axis('yaxis', units='ypxl', data=y, index=0), ])

    viewer.show_data(data_to_plot)
    viewer.setXaxis(Axis(data=x, label='This is x axis', units='au'))
    viewer.setYaxis(Axis(data=y, label='This is y axis', units='au'))
    win.show()
    app.processEvents()
    sys.exit(app.exec_())



from typing import List, Tuple, Any, TYPE_CHECKING, Union
import sys

from qtpy import QtWidgets

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_utils import config as configmod
from pymodaq_gui.utils.utils import start_qapplication


from pymodaq_data.plotting.plotter.plotter import PlotterBase, PlotterFactory
from pymodaq_data.data import DataWithAxes, DataToExport
from pymodaq_gui.plotting.data_viewers import (Viewer1D, Viewer2D, ViewerND, ViewerDispatcher,
                                                 Viewer0D, ViewersEnum)
from pymodaq_gui.plotting.data_viewers.viewer import ViewerBase, viewer_factory
from pymodaq_gui.utils.dock import DockArea

logger = set_logger(get_module_name(__file__))
config = configmod.Config()


@PlotterFactory.register()
class Plotter(PlotterBase):
    backend = 'qt'

    def __init__(self, **_ignored):
        super().__init__()

    def plot(self, data: Union[DataWithAxes, DataToExport], viewer=None, **kwargs) -> ViewerBase:
        """

        Parameters
        ----------
        data
        viewer
        kwargs

        Returns
        -------

        """
        do_exit = False
        qapp = QtWidgets.QApplication.instance()
        if qapp is None:
            do_exit = True
            qapp = start_qapplication()

        if viewer is None:
            if isinstance(data, DataToExport):
                widget = DockArea()
                viewer = ViewerDispatcher(widget, title=data.name)
            else:
                widget = QtWidgets.QWidget()
                viewer_enum = ViewersEnum.get_viewers_enum_from_data(data)
                viewer = viewer_factory.get(viewer_enum.name, parent=widget)
            widget.show()

        if viewer is not None:
            viewer.show_data(data, **kwargs)
            if isinstance(viewer, Viewer1D):
                if not viewer.is_action_checked('errors'):
                    viewer.get_action('errors').trigger()
            QtWidgets.QApplication.processEvents()

        if do_exit:
            sys.exit(qapp.exec())
        return viewer


if __name__ == '__main__':
    from pymodaq.utils import data as data_mod
    import numpy as np
    from pymodaq.utils.math_utils import gauss1D
    from pymodaq.utils.plotting.plotter.plotter import PlotterFactory

    qapp = start_qapplication()

    plotter_factory = PlotterFactory()

    x = np.random.randint(201, size=201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)

    QtWidgets.QApplication.processEvents()
    dwa = data_mod.DataRaw('mydata', data=[y1, y2],
                            axes=[data_mod.Axis('myaxis', 'units', data=x, index=0,
                                                spread_order=0)],
                            nav_indexes=())

    dwa.plot('qt')
    dwa.as_dte('mydte').plot('qt')
    sys.exit(qapp.exec())



from copy import deepcopy

from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph.parametertree.Parameter import ParameterItem
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem
from pyqtgraph.parametertree import Parameter


class ItemSelect_pb(QtWidgets.QWidget):
    def __init__(self,checkbox=False,):

        super(ItemSelect_pb, self).__init__()
        self.initUI(checkbox,)

    def initUI(self, checkbox=False,):        
        #### Widgets ###        
        # ListWidget
        self.itemselect = ItemSelect(checkbox)
        # Pushbutton Add
        self.add_pb = QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)        
        # Pushbutton Remove
        self.remove_pb = QtWidgets.QPushButton()
        self.remove_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/remove.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.remove_pb.setIcon(icon3)               
        #### Layout ###
        self.ver_layout = QtWidgets.QVBoxLayout()    
        self.ver_layout.addWidget(self.add_pb)        
        self.ver_layout.addWidget(self.remove_pb)            
        self.ver_layout.setSpacing(0)
                
        self.hor_layout = QtWidgets.QHBoxLayout()        
        self.hor_layout.addWidget(self.itemselect)
        self.hor_layout.addLayout(self.ver_layout)
        
        self.hor_layout.setSpacing(0)
        self.setLayout(self.hor_layout)


class ItemSelect(QtWidgets.QListWidget):
    def __init__(self, hasCheckbox=True):
        QtWidgets.QListWidget.__init__(self)
        self.hasCheckbox = hasCheckbox # Boolean indicating if listwidget item uses checkbox ot not
        self.selItems = [] # Dummy variable to keep track of click order
        self.itemDoubleClicked.connect(self.doubleClickSelection)
        
    def doubleClickSelection(self,item:QtWidgets.QListWidgetItem):
        """
            Function to select item. The selection depends if the item uses checkbox or not.
        """        
        if self.hasCheckbox:
            item.setCheckState(int(2*bool(not item.checkState())))
            

    def get_value(self):
        """
            Get the dictionnary of values contained in the QtWidget attribute.

            Returns
            -------
            dictionnary
                The dictionnary of all_items compared to the selectedItems.                                
        """
        allitems = [item.text() for item in self.all_items()]
        if self.hasCheckbox:   
            # Clean up list with non existing entries      
            [self.selItems.remove(item) for item in self.selItems if item not in allitems]        
            for item in self.all_items():
                if item.checkState() != 0: # Item is selected
                    if item.text() not in self.selItems: # if item not in list then add it
                        self.selItems.append(item.text())
                else: # Item is not selected
                    if item.text() in self.selItems:  # if item in list then remove it
                        self.selItems.remove(item.text())
            selitems = self.selItems.copy() #need to copy to correctly emit signal when changed
            
            # selitems = [item.text() for item in self.all_items() if item.checkState()!=0]
        else:
            selitems = [item.text() for item in self.selectedItems()]
            
        return dict(all_items=allitems, selected=selitems)

    def all_items(self) -> list:
        """
            Get the all_items list from the self QtWidget attribute.

            Returns
            -------
            list
                The item list.
        """
        return [self.item(ind) for ind in range(self.count())]
    
    def select_item(self, item: QtWidgets.QListWidgetItem, doSelect:bool = False):
        """
            Function to select item. The selection depends if the item uses checkbox or not.
        """        
        if self.hasCheckbox:
            item.setCheckState(int(2*doSelect))  # 2=QtCore.Qt.Checked, 0=QtCore.Qt.Unchecked
        else:
            item.setSelected(doSelect)

    def set_value(self, values: dict):
        """
            Set values to the all_items attributes filtering values by the 'selected' key.

            =============== ============== =======================================
            **Parameters**    **Type**       **Description**
            *values*          dictionnary    the values dictionnary to be setted.
            =============== ============== =======================================
        """
        # Remove values in selected if they do not exist in all
        values = deepcopy(values)
        [values['selected'].remove(value) for value in values['selected'] if value
         not in values['all_items']]
        
        allitems_text = []
        # Check existing items and remove unused ones
        for item in self.all_items():     
            if item.text() not in values['all_items']:  # Remove items from list if text not
                # in values
                item = self.takeItem(self.row(item))
            else:
                allitems_text.append(item.text())  # Add items to list
            self.updateGeometry()
        # Create items if needed
        for value in values['all_items']:  # Loop through all values
            if value not in allitems_text:  # Test if object already exists
                item = QtWidgets.QListWidgetItem(value) # Create object
                if self.hasCheckbox:  # Add checkbox if required
                    item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)      
                    self.select_item(item, doSelect=False)
                    # Make sure item is not selected (checkbox not appearing somehow without)
                self.addItem(item)  # Add object to widget
                 
        allitems = self.all_items()  # All selectable items
        # Selection process
        for item in allitems:
            self.select_item(item, doSelect=False)
        for value in values['selected']:  # Loop through selected to retain selection order
            item = allitems[[item.text() for item in allitems].index(value)]
            self.select_item(item, doSelect=True)
        QtWidgets.QApplication.processEvents()

    def sizeHint(self):
        return QtCore.QSize(super().sizeHint().width(), 25 * self.count())


class ItemSelectParameterItem(WidgetParameterItem):
    
    def makeWidget(self):
        """
            | Make and initialize an instance of ItemSelect_pb with itemselect value.
            | Connect the created object with the plus and minus buttonClicked function.

        """
        self.asSubItem = True
        self.hideWidget = False
        opts = self.param.opts
        
        if 'checkbox' in opts and opts['checkbox']:
            w = ItemSelect_pb(checkbox=opts['checkbox'])
            w.sigChanged = w.itemselect.itemChanged
        else:
            w = ItemSelect_pb()
            w.sigChanged = w.itemselect.itemSelectionChanged

            
        if 'dragdrop' in opts and opts['dragdrop']:        
            w.itemselect.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)

        w.itemselect.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        # w.itemselect.setMinimumHeight(opts.get('min_height', 0))
        # w.itemselect.setMaximumHeight(opts.get('height', 70))

        w.itemselect.setResizeMode(QtWidgets.QListView.Adjust)
        w.add_pb.setVisible(opts.get('show_pb', False))
        w.remove_pb.setVisible(opts.get('show_mb', False))

        if 'tip' in opts:
            w.setToolTip(opts['tip'])
        w.value = w.itemselect.get_value
        w.setValue = w.itemselect.set_value
        w.add_pb.clicked.connect(self.pb_buttonClicked)
        w.remove_pb.clicked.connect(self.mb_buttonClicked)        
        return w

    def pb_buttonClicked(self):
        """
           Append to the param attribute the dictionnary obtained from the QtWidget add parameter procedure.
        """

        text, ok = QtWidgets.QInputDialog.getText(None, "Enter a value to add to the parameter",
                                                  "String value:", QtWidgets.QLineEdit.Normal)
        if text in self.param.value()['all_items']:
            print('Entry already exists, please use a different name.')
            return
        if ok and not (text == ""):
            all = self.param.value()['all_items']
            all.append(text)
            sel = self.param.value()['selected']
            sel.append(text)
            val = dict(all_items=all, selected=sel)
            self.param.setValue(val)
            self.param.sigValueChanged.emit(self.param, val)
            
    def mb_buttonClicked(self):
        """
           Remove the selected Qwidget items by removing the entries in the parameter attribute.
        """                       
        items_to_be_removed = self.widget.itemselect.selectedItems()
        if len(items_to_be_removed) > 0:
            all = self.param.value()['all_items']
            sel = self.param.value()['selected'] 
            for item in items_to_be_removed:
                if item.text() in all:
                    all.remove(item.text())
                    if item.text() in sel:
                        sel.remove(item.text())
            val = dict(all_items=all, selected=sel)
            self.param.setValue(val)
            self.param.sigValueChanged.emit(self.param, val)            

    def optsChanged(self, param, opts):
        """
            Called when any options are changed that are not name, value, default, or limits.

            See Also
            --------
            optsChanged
        """
        # print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)
        
        self.widget.add_pb.setVisible(opts.get('show_pb', False))
        self.widget.remove_pb.setVisible(opts.get('show_mb', False))
        if 'height' in opts:
            self.widget.itemselect.setMaximumHeight(opts['height'])

    def valueChanged(self, param, val, force=False):
        super().valueChanged(param, val, force)
        self.widget.itemselect.updateGeometries()

class ItemSelectParameter(Parameter):
    """
        Editable string; displayed as large text box in the tree.

        =============== ======================================
        **Attributes**    **Type**
        *itemClass*       instance of ItemSelectParameterItem
        *sigActivated*    instance of pyqt Signal
        =============== ======================================
    """
    itemClass = ItemSelectParameterItem
    sigActivated = QtCore.Signal(object)

    def activate(self):
        """
            Activate the "Activated" signal attribute0
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)






# -*- coding: utf-8 -*-
"""
Created the 26/01/2023

@author: Sebastien Weber
"""
from typing import List

from qtpy import QtWidgets, QtCore

from pymodaq_gui.plotting.data_viewers.viewer1Dbasic import Viewer1DBasic
from pymodaq_data.data import Axis
from pymodaq_utils.math_utils import find_index


class AxesViewer(QtCore.QObject):
    navigation_changed = QtCore.Signal()

    def __init__(self, parent_widget: QtWidgets.QWidget):
        super().__init__()
        self._axes: List[Axis]
        self.parent = parent_widget
        self.parent.setLayout(QtWidgets.QVBoxLayout())
        self.nav_axes_viewers: List[Viewer1DBasic] = []

    def clear_viewers(self):
        while len(self.nav_axes_viewers) != 0:
            viewer = self.nav_axes_viewers.pop(0)
            self.parent.layout().removeWidget(viewer.parent)
            viewer.parent.close()

    def add_viewers(self, nviewers: int):
        widgets = []
        for ind in range(nviewers):
            widgets.append(QtWidgets.QWidget())
            self.parent.layout().addWidget(widgets[-1])
            self.nav_axes_viewers.append(Viewer1DBasic(widgets[-1], show_line=True))

    def set_nav_viewers(self, axes: List[Axis]):
        self._axes = axes
        if len(self.nav_axes_viewers) != len(axes):
            self.clear_viewers()
            self.add_viewers(len(axes))

        for ind in range(len(axes)):
            self.nav_axes_viewers[ind].roi_line_signal.connect(self._emit_nav_signal)
            self.nav_axes_viewers[ind].show_data([axes[ind].get_data()])
            self.nav_axes_viewers[ind].set_axis_label(dict(orientation='bottom',
                                                           label='Scan index',
                                                           units=''))
            self.nav_axes_viewers[ind].set_axis_label(dict(orientation='left',
                                                           label=axes[ind].label,
                                                           units=axes[ind].units))

    def _emit_nav_signal(self):
        self.navigation_changed.emit()

    def get_crosshairs(self):
        return tuple([viewer.get_line_position() for viewer in self.nav_axes_viewers])

    def get_indexes(self):
        return [int(cross) for cross in self.get_crosshairs()]

    def setVisible(self, show=True):
        """convenience method to show or hide the paretn widget"""
        self.parent.setVisible(show)


if __name__ == '__main__':
    import sys
    import numpy as np
    app = QtWidgets.QApplication(sys.argv)

    widget = QtWidgets.QWidget()
    prog = AxesViewer(widget)
    widget.show()
    labels = ['']
    N = 2
    axes = [Axis(label=f'Axis{ind:02d}', units='s', data=np.random.rand(50)) for ind in range(N)]
    prog.set_nav_viewers(axes)

    def print_positions():
        print(prog.get_crosshairs())

    prog.navigation_changed.connect(print_positions)

    sys.exit(app.exec_())




from typing import List, Union
import numpy as np

from qtpy import QtWidgets

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_data.data import DataToExport, DataWithAxes, Axis, DataSource

from pymodaq_utils.enums import enum_checker
from pymodaq_utils.factory import ObjectFactory
from pymodaq_gui.plotting import data_viewers
from pymodaq_gui.plotting.data_viewers.base import ViewerBase, ViewersEnum
from pymodaq_gui.utils import DockArea, Dock

config_viewers = {}

logger = set_logger(get_module_name(__file__))


def get_viewer_enum_from_axes(Naxes: int):
    if Naxes < 0:
        raise ValueError('Naxes could not be below 0')
    if Naxes == 0:
        viewer_enum = ViewersEnum['Viewer0D']
    elif Naxes == 1:
        viewer_enum = ViewersEnum['Viewer1D']
    elif Naxes == 2:
        viewer_enum = ViewersEnum['Viewer2D']
    else:
        viewer_enum = ViewersEnum['ViewerND']
    return viewer_enum


class ViewerFactory(ObjectFactory):
    def get(self, viewer_name, **kwargs):
        if viewer_name not in ViewersEnum.names():
            raise ValueError(f'{viewer_name} is not a valid PyMoDAQ Viewer: {ViewersEnum.names()}')
        return self.create(viewer_name, **kwargs)

    @property
    def viewers(self):
        return self.keys


@ViewerFactory.register('Viewer0D')
def create_viewer0D(parent: QtWidgets.QWidget = None, **_ignored):
    return data_viewers.viewer0D.Viewer0D(parent)


@ViewerFactory.register('Viewer1D')
def create_viewer1D(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewer1D.Viewer1D(parent)


@ViewerFactory.register('Viewer2D')
def create_viewer2D(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewer2D.Viewer2D(parent)


@ViewerFactory.register('ViewerND')
def create_viewerND(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewerND.ViewerND(parent)


# @ViewerFactory.register('ViewerSequential')
# def create_viewer_sequential(widget: QtWidgets.QWidget, **_ignored):
#     return data_viewers.viewer_sequential.ViewerSequential(widget)


viewer_factory = ViewerFactory()


class ViewerDispatcher:
    """MixIn class to add easy control for adding multuiple data viewers in docks depending on
    data to be plotted

    Parameters
    ----------
    dockarea: DockArea
    title: str
    next_to_dock: Dock
        (deprecated) has no effect
    direction: str
        either 'right', 'left', 'bottom', 'top'.

    """

    def __init__(self, dockarea: DockArea = None, title: str = '', next_to_dock: Dock = None,
                 direction='right'):
        super().__init__()
        self._title = title

        self._next_to_dock = next_to_dock

        if dockarea is None:
            dockarea = DockArea()
            dockarea.show()
        self.dockarea = dockarea

        self._direction = direction

        self._viewer_docks = []
        self._viewer_widgets = []
        self._viewer_types = []
        self._viewers = []

    @property
    def viewers(self) -> List[ViewerBase]:
        return self._viewers

    @property
    def viewer_docks(self) -> List[Dock]:
        return self._viewer_docks

    @property
    def viewer_widgets(self) -> List[QtWidgets.QWidget]:
        return self._viewer_widgets

    @property
    def viewer_types(self) -> List[ViewersEnum]:
        return self._viewer_types

    def remove_viewers(self, Nviewers_to_leave: int = 0):
        """Remove viewers from the list after index Nviewers_to_leave

        Parameters
        ----------
        Nviewers

        Returns
        -------

        """
        while len(self.viewer_docks) > Nviewers_to_leave:
            widget = self.viewer_widgets.pop()
            widget.close()
            dock = self.viewer_docks.pop()
            dock.close()
            self.viewers.pop()
            self.viewer_types.pop()
            QtWidgets.QApplication.processEvents()

    def add_viewer(self, viewer_type: ViewersEnum, dock_viewer=None, dock_name=None):
        viewer_type = enum_checker(ViewersEnum, viewer_type)

        if dock_viewer is None:
            if dock_name is None:
                dock_name = f'{self._title}_Viewer_{len(self.viewer_docks) + 1}'
            dock_viewer = Dock(dock_name, size=(350, 350), closable=False)
        self.viewer_docks.append(dock_viewer)

        self._viewer_widgets.append(QtWidgets.QWidget())
        self.viewers.append(viewer_factory.get(viewer_type.name, parent=self._viewer_widgets[-1]))

        self.viewer_types.append(viewer_type)

        self.viewer_docks[-1].addWidget(self._viewer_widgets[-1])
        # if len(self.viewer_docks) == 1:
        #     if self._next_to_dock is not None:
        #         self.dockarea.addDock(self.viewer_docks[-1], 'right', self._next_to_dock)
        #     else:
        #         self.dockarea.addDock(self.viewer_docks[-1])
        # else:
        #     self.dockarea.addDock(self.viewer_docks[-1], 'right', self.viewer_docks[-2])
        self.dockarea.addDock(self.viewer_docks[-1], self._direction)

    def update_viewers(self, viewers_type: List[Union[str, ViewersEnum]],
                       viewers_name: List[str] = None, force=False):
        """

        Parameters
        ----------
        viewers_type: List[ViewersEnum]
        viewers_name: List[str] or None
        force: bool
            if True remove all viewers before update else check if new viewers type are compatible with old ones

        Returns
        -------

        """

        Nviewers_to_leave = 0
        if not force:
            # check if viewers are compatible with new data dim
            for ind, viewer_type in enumerate(viewers_type):
                if len(self.viewer_types) > ind:
                    if viewer_type == self.viewer_types[ind]:
                        Nviewers_to_leave += 1
                    else:
                        break
                else:
                    break
        self.remove_viewers(Nviewers_to_leave)
        ind_loop = 0
        while len(self.viewers) < len(viewers_type):
            self.add_viewer(viewers_type[Nviewers_to_leave + ind_loop],
                            dock_name=viewers_name[Nviewers_to_leave + ind_loop]
                            if viewers_name is not None else None)
            ind_loop += 1
        QtWidgets.QApplication.processEvents()

    def close(self):
        for dock in self.viewer_docks:
            dock.close()

    def show_data(self, data: DataToExport, **kwargs):
        """ Convenience method. Display each dwa in a dedicated data viewer"""
        viewer_types = [ViewersEnum.get_viewers_enum_from_data(dwa) for dwa in data]
        viewer_names = [dwa.name for dwa in data]
        if self.viewer_types != viewer_types:
            self.update_viewers(viewer_types, viewer_names)
        for viewer, dwa in zip(self.viewers, data):
            if len(dwa.axes) != len(dwa.shape):
                dwa.create_missing_axes()
            viewer.show_data(dwa)


if __name__ == '__main__':

    LABEL = 'A Label'
    UNITS = 'units'
    OFFSET = -20.4
    SCALING = 0.22
    SIZE = 20
    DATA = OFFSET + SCALING * np.linspace(0, SIZE - 1, SIZE)

    DATA0D = np.array([2.7])
    DATA1D = np.arange(0, 10)
    DATA2D = np.arange(0, 5 * 6).reshape((5, 6))
    DATAND = np.arange(0, 5 * 6 * 3).reshape((5, 6, 3))
    Nn0 = 10
    Nn1 = 5


    def init_axis(data=None, index=0):
        if data is None:
            data = DATA
        return Axis(label=LABEL, units=UNITS, data=data, index=index)


    def init_data(data=None, Ndata=1, axes=[], name='myData', source=DataSource['raw'],
                  labels=None) -> DataWithAxes:
        if data is None:
            data = DATA2D
        return DataWithAxes(name, source, data=[data for ind in range(Ndata)],
                            axes=axes, labels=labels)


    def ini_data_to_export():
        dat1 = init_data(data=DATA2D, Ndata=2, name='data2D')
        dat2 = init_data(data=DATA1D, Ndata=3, name='data1D')
        data = DataToExport(name='toexport', data=[dat1, dat2])
        return dat1, dat2, data

    import sys

    app = QtWidgets.QApplication(sys.argv)

    dockarea = DockArea()
    prog = ViewerDispatcher(dockarea=dockarea, title='Dispatcher')
    dockarea.show()

    _, _, dte = ini_data_to_export()

    # N = 2
    # viewers = ['Viewer0D', 'Viewer1D', 'Viewer2D']
    # viewers = [ViewersEnum[random.choice(viewers)] for ind in range(N)]
    # viewers = ['Viewer2D', 'Viewer2D', ]
    # print(viewers)
    # prog.update_viewers(viewers)

    prog.show_data(dte)

    sys.exit(app.exec_())
